;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: ADB; -*- 

(in-package 'adb)

;;;;Making and processing assertions
;;
;; Defines the forms of facts and what they turn into.
;;
;;A wrinkle which wasn't a problem in DEBACLE concerns the mapping from
;;forms to assertion classes.  DEBACLE assumed the CAR was always
;;the right thing, and if the datum was to be indexed under multiple names
;;(to speed access, for instance) the variant mechanism allowed the same assertion
;;to be the referent of several forms.  But a standard template for assertions when
;;thinking about the ATMS is 
;		((<function> .  <arguments>) . <values>)
;Sometimes you want to get the list of values, in which case the DEBACLE solution is right.
;But for the rules that define the function (or predicates and relations, in the special
;case where <values> = :TRUE, :FALSE) it is the CAAR of the form that is desired.
;    The alternative we shall use here is to always index under the leftmost symbol in the
;tree formed by the expression.  This means that more pattern-matches will fail, since things
; haven't been divided as narrowly as they might otherwise be.  But if this turns out to be
; a problem it can always be changed.

(export '(assertq assert! oneof pairwise-nogoods taxonomy
	  at-least-one at-most-one assumep create-var assume-var nogood justify
	  deffact deffact-e defassumption defassumption-e defassumep defassumep-e
	  assume assume-if-needed defassert defFunct *justify*))

(defvar *justify* 0) ;; Number of justifications created.
(init-form (setq *justify* 0))

(defun function-term? (form) (not (null (cdr (last form)))))

;;; Making assertions and running rules

(defmacro assertq (fact) `(assert1 (desugar-pattern ',fact)))

(defun assert! (fact)  (assert1 (desugar-pattern fact)))

;;; Instead of screwing around with packages directly,
;;; just re-define these.  

(defmacro user::assertq (fact) `(assert1 (desugar-pattern ',fact)))

(defun user::assert! (fact) (assert1 (desugar-pattern fact)))

(defun assert1 (fact &aux node)
  ;; FACT must be a proposition or ground term.
  ;; If you want to assert connectives, use CONSTRAINT.
  (setq node (get-tms-node fact))
  (incf *justify*)
  (tms::justify-node node '(ASSERTION))
  node)

(defmacro assumeq (fact) `(assume1 (desugar-pattern ',fact)))
(defun assume! (fact) (assume1 (desugar-pattern fact)))
(defmacro user::assumeq (fact) `(assume1 (desugar-pattern ',fact)))
(defun user::assume! (fact) (assume1 (desugar-pattern fact)))

(defmacro assume (fact)
  (format t "Warning: You are using ADB::ASSUME, which is obsolete.  This will not work in the future.")
  `(assume! ,fact))

(defun assume1 (fact &aux node)
  ;; FACT must be a proposition or ground term.
  (setq node (get-tms-node fact))
  (tms::assume-node node)
  node)

(defun assume-if-needed (form)
  (cond ((fact-assumption? form) (fact-assumption form))
	(t (assume! form) (fact-assumption form))))

(defun assumep (base-fact &aux true-node false-node tms-class)
  ;; Johan twiddled this for new ATMS.
  ;; "Assume Proposition" -- makes :TRUE,  :FALSE versions of a fact and
  ;; makes them a closed, exclusive class. 
  (setq true-node (get-tms-node (cons base-fact ':TRUE))
	false-node (get-tms-node (cons base-fact ':FALSE))
	tms-class (car (tms::node-classes true-node)))
  (tms::assume-symbol true-node false-node tms-class))

(defmacro DefFact (fact)
  `(progn (cond ((okay-packet? *default-packet*)
		 (add-to-packet *default-packet* '(assert! ',fact)))
		(t (assert! ',fact)))))

(defmacro DefAssumption (fact)
  `(progn (cond ((okay-packet? *default-packet*)
		 (add-to-packet *default-packet* '(assume! ',fact)))
		(t (assume!! ',fact)))))

;;;; Additional logical constraints

;;; Johan: Temporary while we hack the new interface.
;;  KDF: Should check to see if TRUE/FALSE pairs, since these
;;       are supported internally now. 

(defun oneof (forms &aux nodes)
  (dolist (f forms)
    (push (get-tms-node f) nodes))
  (tms::assume-oneof nodes (cons 'ONEOF nodes)))

(defun pairwise-nogoods (forms &optional (informant 'pairwise-nogoods))
  (do ((this (car forms) (car rest)) ;Install pairwise nogoods.
       (rest (cdr forms) (cdr rest)))
      ((null this) forms)
    (dolist (other rest)
;      (print (list this other))
      (nogood (list this other) informant))))

(defun taxonomy (forms)
  ;;Assumes forms are propositions, which already have TMS nodes
  (at-most-one forms)
  (at-least-one forms))

(defun at-least-one (forms)
  ;;;Two ways to do this: Either just state that all false is
  ;;;nogood, or install justifications for Nth if N-1 are false.
  ;;; The former is more efficient if the forms are going to be
  ;;; a component of a solution space.  The latter is necessary
  ;;; if they aren't.  Since the former is easy, this code does
  ;;; the latter.
  (cond (*new* ;; Johan:
	 (clause (mapcar #'(lambda (f) (cons f :TRUE)) forms) nil 'AT-LEAST-ONE))
	(t (dolist (form forms)
	     (justify (cons form ':TRUE)
		      (mapcar #'(lambda (f) (cons f ':FALSE))
			      (remove form forms))
		      'AT-LEAST-ONE)))))

(defun at-most-one (forms)
  ;; If any one of them is true, then the others must be false.
  (cond (*new*   ;; Johan installed this
	 (do ((form forms (cdr forms)))
	     ((null (cdr forms)))
	   (dolist (other-form (cdr forms))
	     (clause (list (cons (car form) :TRUE)
			   (cons other-form :TRUE)) nil 'AT-MOST-ONE))))
  (t
  (dolist (form forms)
    (dolist (other-form (remove form forms))
      (justify (cons other-form ':FALSE)
	       (list (cons form ':TRUE)) 'AT-MOST-ONE))))))

(defun at-most-one-via-nogoods (forms)
  (pairwise-nogoods (mapcar #'(lambda (x) (cons x ':TRUE))
			    forms) 'AT-MOST-ONE))



;;;; Installing nogoods and justifications

(proclaim '(special *nogoods*)) ;register for debugging

(defun nogood (forms &optional (informant 'ADB-NOGOOD))
  (if-debug-flag nogoods
      (format t "~%Adding nogood:~% ~A."
	      (lprinc-string (cons 'AND forms))))
  (if-debug-flag record-nogoods
   (cond ((boundp '*nogoods*) (setf *nogoods*
				    (list (mapcar #'get-tms-node forms))))
	 (t (setq *nogoods* (list (mapcar #'get-tms-node forms))))))	
  (incf *justify*)
  (tms::justify-node tms::*contra-node*
		    (cons informant
			  (mapcar #'get-tms-node forms))))

(defun invert-fact-if-needed (f)
  (if (eq (car f) ':NOT)
      (if (eq (cdadr f) ':TRUE)
	  (cons (caadr f) ':FALSE)
	  (if (eq (cdadr f) ':FALSE)
	      (cons (caadr f) ':TRUE)
	      (error "~%Improper use of negation on ~A." f)))
      f))

(defun justify (consequent antecedents &optional (informant 'ADB-JUSTIFICATION))
  ;essentially P1&P2&...Pn->Q where P1,...Pn are ground terms
  ;This is needed because of gaping hole in translation theory.
  ;maybe fixable, maybe not.  Sigh.
  (if-debug-flag justifications
   (format t "~%Justifying ~A via ~A on~% ~A." (lprinc-string consequent)
	   informant
	   (lprinc-string (cons 'and antecedents))))
  (incf *justify*)
  (cond (*new* ;;Johan.
	 (tms::clause informant
		      (list (get-tms-node consequent))
		      (mapcar #'(lambda (f) (get-tms-node f)) antecedents)
		      ))
	(t (tms::justify-node (get-tms-node consequent)
			      (cons informant
				    (mapcar #'(lambda (f) (get-tms-node f)) antecedents))))))

;;;; Variants of justify

;;; Johan.
(defun clause (positives negatives &optional (informant 'ADB-JUSTIFICATION) &aux ps ns)
  (dolist (p positives) (push (get-tms-node p) ps))
  (dolist (n negatives) (push (get-tms-node n) ns))
  (tms::clause informant ps ns))

(defun justify-w-nodes ;; assumes the arguments are already TMS nodes
       (consequent antecedents &optional (informant 'ADB-Node-Justify))
  (incf *justify*)
  (if-debug-flag node-justify 
    (format t "~% NJ: ~A: ~A" informant
	    (lprinc-string `(-> (AND ,@ 
				  (mapcar #'view-node antecedents))
				,(view-node consequent)))))
  (tms::justify-node consequent (cons informant antecedents)))

(defun nogood-w-nodes (nodes &optional (informant :Nogood-nodes))
  (if-debug-flag node-nogoods 
    (format t "~% NN: ~A: ~A" informant
	    (lprinc-string `(not (AND ,@(mapcar #'view-node nodes))))))
  (incf *justify*)
  (tms::justify-node tms::*contra-node* (cons informant nodes)))
