;;; -*- Mode: LISP; Package: ADB; Syntax: Common-lisp -*-

(in-package 'adb)

;;;; ATMoSphere clause manipulation facilities

;; This stuff may change considerably once we're comfortable with the DNF-clause
;; justifications in the ATMS.  For now, horn clauses are assumed as the target
;; description.

(export '(justify-facts-with-connectives flip-truths))

(defun justify-facts-with-connectives (fact antecedents
					    &optional (informant 'Syntaxer))
  ;;; Produces forms which, at evaluation time, will result in the "right" justifications.
  ;;; Currently handles the following cases:
  ;;; plain facts -- produces a single adb::rjustify statement.
  ;;; (-> (and nantecedents) <fact>)
  ;;; Produces justification with nantecedents U antecedents.
  ;;; (and . <facts>)
  ;;;  Produces justifications for each individual fact.
  ;;; Recursion is allowed, but not encouraged.  For example,
  ;;; extending this stuff to include OR and NOT would be a pain.
  ;;; The problem is that connectives must be treated differently when
  ;;; appearing in the antecedent from appearing in the consequent.
  (cond ((not (listp fact)) `((adb::rjustify ,fact ,antecedents ',informant)))
	(t (case (car fact)
	     (IFF (nconc (justify-facts-with-connectives
			   `(implies ,(cadr fact) ,(caddr fact))
			   antecedents informant)
			 (justify-facts-with-connectives
			   `(implies ,(caddr fact) ,(cadr fact))
			   antecedents informant)
			 ;; These ensure that when one is known to be false, the other is
			 ;; as well.
			 (justify-facts-with-connectives
			   `(implies ,(flip-truths (cadr fact)) ,(flip-truths (caddr fact)))
			   antecedents informant)
			 (justify-facts-with-connectives
			   `(implies ,(flip-truths (caddr fact)) ,(flip-truths (cadr fact)))
			   antecedents informant)))			 
	     (IMPLIES 
	       (cond ((and (listp (cadr fact))
			   (eq (caadr fact) 'AND))
		      (justify-facts-with-connectives
			(caddr fact)
			;; Slightly more expensive, but makes justifications easier for
			;; user to recognize.  It's done at compile-time, anyway.
			(adb::order-preserving-union (cdadr fact)
						     antecedents :test #'equal)))
		     (t (justify-facts-with-connectives
			  (caddr fact)
			  (if (member (cadr fact) antecedents
				      :test #'equal)
			      antecedents
			      (cons (cadr fact) antecedents))))))
	     (AND (mapcan #'(lambda (conse)
			      (justify-facts-with-connectives
				conse antecedents)) (cdr fact)))
	     (t `((adb::rjustify ,fact ,antecedents ',informant)))))))

;;;; Negating expressions.

(defvar *flip-truth-value* nil)

(proclaim '(special *flip-truth-value* *quantifiers*))

(defun flip-truths (x)
  (cond ((null x) nil)
	((not (listp x)) x)
	((eq (cdr x) ':TRUE) (cons (flip-truths (car x)) ':FALSE))
	((eq (cdr x) ':FALSE) (cons (flip-truths (car x)) ':TRUE))
	((eq (car x) ':EVAL)
	 ;; Should include an unwind-protect, as per page 141 of the CL manual
	 `(:Eval (let ((*old-flip-truth-value*
			 *flip-truth-value*))
		   (unwind-protect 
		       (progn (setq *flip-truth-value* 
				    (not *flip-truth-value*))
			      ,(flip-truths (cadr x)))
		     (setq *flip-truth-value* *old-flip-truth-value*)))))
	((member (car x) adb::*quantifiers*) x)
	(t (cons (flip-truths (car x)) (flip-truths (cdr x))))))

