;;; -*- Mode: LISP; Syntax: Common-lisp; Package: ADB -*-

(in-package 'ADB)

;;;; Copyright (c) 1989, Kenneth D. Forbus, University of Illinois
;;;; All Rights Reserved.

;;;; Closing a fact
;; This is a form of negation by failure.
;; The idea is that at some stage you presume you know all the circumstances in which
;; a fact holds, and you want to ensure that for any other circumstances, its negation
;; will hold.  Given a fact F, there are two forms of this:
;; 1. Close over F's justifications.  Namely, ensure that if you do not believe any of the
;;    facts which comprise a consistent set of antecedents for one of F's justifications.
;; 2. Close over F's label. Namely, ensure that if none of the elements of F's label
;;    are believed, then the negation will be believed.   
;; Both of these are implemented by creating justifications for (not F) based on
;;  its antecedents, either by examining the label or the justifications leading to it.
;; Each scheme has its advantages and disadvantages.
;; In most reasonable circumstances these two schemes will be equivalent. But not always.
;; Closing over justifications presumes that the antecedent facts themselves are closed.
;;   This is certainly the case for antecedents which are assumptions (assuming all
;;   assumptions are propositional), so closing over labels satisfies
;;   this presumption trivially 
;; Closing over labels presumes that the label of F is maximal, that is, no new consistent
;;   environment will ever be added to its label.  If a system is making no more assumptions,
;;   or one can guarentee that new assumptions won't propagate back to contribute new
;;   environments to F's label, then closing over labels is safe.  Otherwise, environments
;;   which should be consistent may be ruled out. 

;; Both methods are defined below. 

(defvar *close-default* :by-justifications)

(defun close-fact (form value)
  (if-debug-flag close
		 (format t "~%   (Closing ~A as ~A.)" form value))
  (if (eq *close-default* :by-label) (close-fact-by-label form value)
      (close-fact-by-justifications form value)))

(defun close-fact-by-label (form value &aux pform plabel antes neg nform nlabel)
  (unless (or (eq value :TRUE) (eq value :FALSE))
    (error "Can only close standard propositions: ~A=~A." form value))
  (setq pform (cons form value)
	neg (if (eq value :TRUE) :FALSE :TRUE)
	nform (cons form neg)
	plabel (environments pform))
  (cond ((null plabel) ;; Never true
	 (cond ((setq nlabel (environments nform)) ;; negation might hold
		(close-fact-by-label form neg))
	       (t (assert! nform))))
	((eq (car plabel) tms::*empty-env*)) ;; Always true
	(t ;; Somewhere in between.
	 ;; Empirical question: is it worth doing a bit of resolution on
	 ;; the label?  Assume for now no.
	 (setq antes (mapcar #'environment-facts plabel))
	 (dolist (nantes (simplify-clauses (compute-negated-antecedents antes)))
	   (justify nform nantes :Negation-by-maximal-assumption)))))

;;;; Clause simplification
;; Worth doing, since the number of justifications can sometimes be greatly reduced.

;; This is adapted from the BPS LTMS code.

(defun simplify-clauses (cset)
  ;; Two kinds of simplifications: 1. One clause might subsume another.
  ;; 2. Two clauses might be resolvable.
  ;; Do the subsumption checks first to minimize effort.
  ;; In fact, for now punt on resolution.
  (dolist (clause cset)
    (when clause 
      (mapl #'(lambda (rest)
		(unless (and (car rest) (eq clause (car rest))) ;; skip yourself
		  (when (clause-subsumed-by? (car rest) clause)
		    ;; clause is smaller, flush other
		    (setf (car rest) nil)))) cset)))
  (setq cset (delete nil cset)))

(defun clause-subsumed-by? (bigger smaller)
  (subsetp smaller bigger :test #'equal))

(defun compute-negated-antecedents (ante-sets &aux negs)
  (labels ((find-negated-antes (flipped-sets current)
	     (cond ((null flipped-sets) 
		    (when current (push current negs)))
		   (t (dolist (el (car flipped-sets))
			(let ((other (assoc (car el) current
					    :test #'equal)))
			  (cond (other
				 (when (eq (cdr el) (cdr other)) ;redundant, go on
				   (find-negated-antes (cdr flipped-sets)
						       current)))
				(t (find-negated-antes (cdr flipped-sets)
						       (cons el current))))))))))
    (find-negated-antes (mapcar #'flip-truths ante-sets) nil))
  negs)

;;;; Closing over justifications

(defun close-fact-by-justifications (form value &aux pform pjusts antes neg nform)
  (unless (or (eq value :TRUE) (eq value :FALSE))
    (error "Can only close standard propositions: ~A=~A." form value))
  (setq pform (cons form value)
	neg (if (eq value :TRUE) :FALSE :TRUE)
	nform (cons form neg))
  ;; Check for special cases: it might be known true or false
  (cond ((member (fact-status pform) '(:TRUE :FALSE)) nil)
	((null (setq pjusts (tms::node-justifications (get-tms-node pform))))
	 (cond ((setq pjusts (tms::node-justifications (get-tms-node nform)))
		(close-fact-by-justifications form neg))
	       (t (assert! nform))))
	(t (setq antes (mapcar #'(lambda (just)
				   (mapcar #'view-node (cdr just))) pjusts))
	   (dolist (nantes (simplify-clauses (compute-negated-antecedents antes)))
	     (justify nform nantes :Negation-by-maximal-justifications)))))

;;; Test cases for closing code

(defun test-close-by-label ()
  (adb::init)
  (assumep 'a)
  (assumep 'b)
  (assumep 'c)
  (assumep 'd)
  (justify '(conc . :TRUE) '((a . :TRUE)(b . :TRUE)) :nbma-test)
  (justify '(conc . :TRUE) '((a . :TRUE)(b . :FALSE)) :nbma-test)
  (justify '(conc . :TRUE) '((c . :TRUE)(d . :FALSE)) :nbma-test)
  (close-fact-by-label 'conc :TRUE)
  (show-facts))

(defun test-close-by-justifications ()
  (adb::init)
  (assumep 'a)
  (assumep 'b)
  (assumep 'c)
  (assumep 'd)
  (justify '(conc . :TRUE) '((a . :TRUE)(b . :TRUE)) :nbma-test)
  (justify '(conc . :TRUE) '((a . :TRUE)(b . :FALSE)) :nbma-test)
  (justify '(conc . :TRUE) '((c . :TRUE)(d . :FALSE)) :nbma-test)
  (close-fact-by-justifications 'conc :TRUE)
  (show-facts))
