;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: ADB; -*-

(in-package 'adb)

;;;; Debacle-like database 

(export '(get-assert-prop put-assert-prop view-node view-class view-assumption
	  referent get-tms-node get-class-tms-node show-data show-classes))

;;; Property-list access for assertions.

(defun get-assert-prop (assert prop) 
  (getf (assert-plist assert) prop))

(defun put-assert-prop (assert val prop)
  (setf (getf (assert-plist assert) prop) val))

;;;; Mapping from asserts to lisp forms.  

(defun view-node (node &aux datum)
  ;; 9/28/89 -- changed to accomodate pnodes, nnodes.
  (cond ((not (tms::careful-node? node)) node)
	((assertion? (setq datum (tms::node-datum node)))
	 (lisp-form (tms::node-datum node)))
	((listp datum) ;; could be negation
	 (cond ((and (eq (car datum) 'not)
		     (assertion? (cdr datum)))
		(list 'NOT (lisp-form (cdr datum))))
	       (t (error "TMS node datum of unexpected form: ~A." node))))
	((stringp datum) datum)  ;; as in "A contradition".  Hopefully no others!
	(t (error "TMS node datum of VERY unexpected form: ~A." node))))

(defun view-class (class) (tms::class-datum class))

(defun get-assumption-node (as) (tms::get-assumption-node as))

(defun view-assumption (as) (view-node (get-assumption-node as)))

(defun node-string-function (node &aux class datum)
  (cond ((assertion? (setq datum (tms::node-datum node)))
	 (format nil "~A=~A" (car (lisp-form datum)) (cdr (lisp-form datum))))
	(t (setq class (car (tms::node-classes node)))
	   (format nil "~A=~A" (if class (tms::class-datum class) "???")
		   (tms::node-datum node)))))

(defvar tms::*node-string* (function node-string-function)
  "Making use of ATMS hook")

;;;; Mapping expressions to assertions and classes 

;; Referent maps from expressions to assertions

(defun referent (form &optional (virtual? NIL))
(cond ((null form)
       (error  ;;; "~%A null value will be returned." 
	       "~%NIL invalid lisp form for datum."))
      ((or (atom form) (listp form) (numberp form))
       (let ((entry (gethash form *fact-hash-table*))) 
       (cond (entry entry)
	     (virtual? (build-assertion form))
	     (t NIL))))
      (t (error "~A Bad format for assertion - REFERENT" form))))

(defun get-assertion-class (form)
  ;;Some expressions will have lists as their CAR's, so recurse until
  ;;non-list object.  A way to minimize the number of calls to unify
  (cond ((null form) (error "~% NIL invalid ADB assertion!"))
	((listp form) (get-assertion-class (car form)))
	((symbolp form) form)
	(t (error "~%Bad class type for assertion: ~A." form))))

(declare (special *number-of-classes*))

(defun lookup-class (form &optional (virtual? nil))
  (cond ((null form) (error "NIL is a bad class name - LOOKUP-CLASS" form))
	((not (symbolp form))
	 (error "Class names must be symbols - LOOKUP-class~%~A" form))  
	(t (let ((loc (gethash form *class-hash-table*)))
	     (cond ((class? loc) loc)
		   (virtual? (setq loc (make-class :unique (incf *class*)
						   :name form))
			     (setf (gethash form *class-hash-table*) loc)
			     loc)
		   (t nil))))))

(defun build-assertion (form)
  (let* ((class (lookup-class (get-assertion-class form) T))
	 (as (make-assert :unique (incf *assert*) :lisp-form form
			  ;; (copy-tree form)
			  ;; NOTE: A small overhead in copying here means levels above
			  ;; can be as filthy as they like...
			  ;; But it 
			  :assertion-class class)))
    (setf (gethash form *fact-hash-table*) as)
    (push as (class-instances class))
    (dolist (rule-instance (class-rules class))
      (try-rule-on rule-instance as))
    as))

;;;; Mapping between data and ATMS nodes.

;;; ****** Can the underlying data model be simplified now that the ATMS
;;; ****** has been cleaned up??

(defun get-tms-node (as)
  (cond ((assertion? as) (get-tms-node1 as))
	((tms::careful-node? as) as)
	(t (get-tms-node1 (referent as T)))))

;;; Johan:
;;; A temporary kludge for the moment.
(defun get-tms-node1 (as)
  (cond ((truth-node as) (truth-node as))
	(t (let ((lform (lisp-form as)))
	     (cond ((listp lform)
		    (let* ((variable (car lform))
			   (value (cdr lform))
			   (adb-class (lookup-class (get-assertion-class lform) t))
			   (class (get-class-tms-node adb-class variable))
			   (node nil)
			   (nnode nil)
			   (nas nil))
			   ;;**** CLASS is not used anymore by ATMS for symbols, does ADDB
			   ;; need it?????
		      (cond ;; If the value is not :TRUE/:FALSE just create the symbol,
			    ;; positively and negatively.
			    ((not (tms::memq value '(:TRUE :FALSE)))
			     (setq node (tms::create-symbol as class)))
			    ;; If the value is :TRUE/:FALSE.  There are two possibilities:
			    ;; [1] the symbol has been created with its two nodes, but
			    ;;     but as has not been hooked up yet.
			    ((and (setq nas
					(referent (cons variable
							(if (eq value :FALSE) :TRUE :FALSE))))
				  (setq nnode (truth-node nas)))
			     (setq node (tms::n-a-neg nnode))
			     (setf (tms::node-datum node) as))
			    ;; [2] The symbol doesn't exist.
			    (t (setq node (tms::create-symbol as))))
		      (setf (truth-node as) node)
		      node))
		 (t (setf (truth-node as) (tms::create-symbol lform))))))))

;;; Johan:
(defun get-class-tms-node (cl variable)
  ;No tms::class should be built which will eventually correspond to
  ;an ADB class except through this function!
  (let ((pair (assoc variable (class-tms-node cl) :test #'equal)))
    (unless pair
      (push (setq pair (cons variable (if *new*
					  (tms::create-class variable T T)
					  (tms::create-class variable))))
	    (class-tms-node cl)))
    (cdr pair)))

;;;; Manipulating the tables 

(defvar *fact-table-size* 4095.)
(defvar *class-table-size* 4095.)

(defun clear-tables (&optional (reset? nil))
  (when (boundp '*fact-hash-table*) (clrhash *fact-hash-table*))
  (if (or reset? (not (boundp '*fact-hash-table*)))
      (setq *fact-hash-table* (make-hash-table :size *fact-table-size*
					       :test #'equal)))
  (when (boundp '*class-hash-table*) (clrhash *class-hash-table*))
  (if (or reset? (not (boundp '*class-hash-table*)))
      (setq *class-hash-table* (make-hash-table :size *class-table-size*
					       :test #'equal))))

(defun list-facts ()
  (let ((facts nil))
    (maphash #'(lambda (key item) ;; something like this
		 (declare (ignore key))
		 (push item facts)) *fact-hash-table*)
    facts))

(defun list-classes ()
  (let ((classes nil))
    (maphash #'(lambda (key item) ;; something like this
		 (declare (ignore key))
		 (push item classes)) *class-hash-table*)
    classes))

;;;; Inspecting and dumping database for debugging
;;;
;;; These are mainly intended for trivial test cases or as
;;; the method of last resort for anything realistic.
;;; They produce ALOT of stuff on a real problem!
;;; More useful debugging tools appear elsewhere.

;; ********* YOU ARE HERE -- make a version of this for LATEX.  Use lprint?? 

(defun core-dump (fname &optional (latex? nil))
  (with-open-file (str fname ':direction ':out)
    (if latex? (format str "\\documentstyle[11pt,alltt]{article}
\\pagestyle{myheadings}
\\makeatletter
\\newcommand{\\singlespacing}{\\let\\CS=\\@currsize\\renewcommand{\\baselinestretch}{0.5}\\small\\CS}
\\newcommand{\\normalspacing}{\\singlespacing}
\\newcommand{\\border}{\\rule{\\textwidth}{0.04in}}
\\makeatother
\\textheight 720pt
\\textwidth 500pt
\\topmargin -0.5in
\\oddsidemargin -0.5in
\\evensidemargin -0.5in
\\begin{document}
\\singlespacing
"))
    (let ((*report-mode* (if latex? ':LATEX ':ASCII)))
      (tms::walk-nodes
	#'(lambda (n)
	    (let ((*standard-output* str))
	      (why n)
	      (why-not n)
	      (format t "~%----------")))))
    (if latex? (format str "~%\\end{document}~%"))))

(defun show-facts (&optional (stream *standard-output*))
  (dolist (as (list-facts))
    (let ((node (get-tms-node as)))
    (format stream "~% ~A : ~A, ~A." (lprinc-string (lisp-form as))
	    (tms::node-status node)
	    (tms::label node)))))

(defun show-classes (&optional (stream *standard-output*))
  (format stream "~%Class name          # facts # rules")
  (dolist (cl (list-classes))
    (format stream "~%~20<~A~>~5<~D~>~5<~D~>" 
	    (class-name cl) (length (class-instances cl))
	    (length (class-rules cl)))))

(defun show-class (cname &key (forms nil) (rules nil) &aux cl)
  (cond ((class? cname) (setq cl cname))
	(t (setq cl (lookup-class cname))))
  (when cl
    (format t "~%Class ~A:" (class-name cl))
    (format t "~% ~D rules, ~D instances." 
	    (length (class-rules cl)) (length (class-instances cl)))
;    (format t "~% Properties: ~A" (class-plist cl))
    (when (and forms (class-instances cl))
      (format t "~% Forms:")
      (dolist (as (class-instances cl))
	(format t "~%   ~A" (lprinc-string (lisp-form as)))))
    (when (and rules (class-rules cl))
      (format t "~% Rules:")
      (dolist (rule (class-rules cl))
        (pprint rule)))))
