;;; -*- Package: ADB; Syntax: Common-lisp; Mode: LISP -*-

(in-package 'ADB)

;;;; Tracing down the killers of an environment

(defun n-1-consistent? (fact-list &aux envs e)
  (dolist (f fact-list)
    (format t "~% Removing ~A -- ~A."
	    f (setq e (adb:create-environment (remove f fact-list))))
    (if (and e (not (tms::env-contradictory e))) (push e envs)))
  envs)

(defun n-2-consistent? (fact-list &aux envs e)
  (dolist (f1 fact-list)
    (dolist (f2 fact-list)
      (unless (eq f1 f2)
	(setq e (adb:create-environment (remove f1 (remove f2 fact-list))))
	(if (and e (not (tms::env-contradictory e))) (pushnew e envs)))))
  envs)

(defun check-pairs (fact-list &aux envs e)
  (do ((f1 (car fact-list) (car fl1))
       (fl1 (cdr fact-list) (cdr fl1)))
      ((null f1) envs)
    (do ((f2 (car fl1) (car fl2))
	 (fl2 (cdr fl1) (cdr fl2)))
	((null f2))
      (unless (eq f1 f2)
	(setq e (adb:create-environment (list f1 f2)))
	(format t "~% ~A & ~A -- ~A." (lprinc-string f1) (lprinc-string f2) e)
	(when (and e (not (tms::env-contradictory e)))
	  (pushnew e envs))))))


(defun check-triples (fact-list &aux envs e losers)
  (do ((f1 (car fact-list) (car fl1))
       (fl1 (cdr fact-list) (cdr fl1)))
      ((null f1) (values envs losers))
    (do ((f2 (car fl1) (car fl2))
	 (fl2 (cdr fl1) (cdr fl2)))
	((null f2))
      (unless (eq f1 f2)
	(do ((f3 (car fl2) (car fl3))
	     (fl3 (cdr fl2) (cdr fl3)))
	    ((null f3))
	  (unless (or (eq f3 f1) (eq f3 f2))
	    (setq e (adb:create-environment (list f1 f2 f3)))
	    (format t "~% ~A~% &~A~% & ~A -- ~A." 
		    (lprinc-string f1) (lprinc-string f2) (lprinc-string f3) e)
	(cond ((and e (not (tms::env-contradictory e)))
	       (pushnew e envs))
	      (t (pushnew (list f1 f2 f3) losers :test #'(lambda (x y)
							   (same-elements? x y)))))))))))

(defun track-down-killer (fact-list)
  ;; here a "killer" is a justification for false that wipes out
  ;; the input fact qua environment.
  ;; Killers can be found by checking the labels of the
  ;; antecedents of the justifications, seeing if one can
  ;; find elements which, when taken together, would yield
  ;; the target environment in union-envs.
  (let ((suspects nil)
	(asn-list (mapcar #'adb:fact-assumption fact-list))
	(count 0))
    (dolist (just (tms::node-justifications tms::*contra-node*))
      (incf count)
      (format t "~% Trying ~A, length ~A." (car just) (length (cdr just)))
      (when (catch 'gotcha (check-just-means (cdr just) asn-list (cdr just)))
	(push just suspects)))
    (format t "~% ~D justifications checked." count)
    suspects))

(defun check-just-means (nodes assumptions nodes-so-far) ;; If run out at same time, win.
  (cond ((null nodes) (if (null assumptions) (throw 'gotcha t)))
	((null assumptions) nil) ;; ran out of assumptions too soon.
	((null (tms::label (car nodes))) ;; empty label already -- recurse!
	 (dolist (just (tms::node-justifications (car nodes)))
	   (unless (intersection (cdr just) nodes-so-far)
	     (check-just-means (append (cdr just) (cdr nodes)) assumptions
			       (append (cdr just) nodes-so-far)))))
	(t ;; check over each env in the label
	 (format t "~%   Label = ~A." (tms::label (car nodes))) 
	 (dolist (env (tms::label (car nodes)))
	   (unless (set-difference (tms::env-assumptions env) assumptions)
	     ;; This env is a subset of the assumptions, so try it out.
	     (check-just-means (cdr nodes)
			       (set-difference assumptions
					       (tms::env-assumptions env))
			       nodes-so-far))))))

;;;;  Some ATMS-level tools

(defun get-node (num)
  (tms::walk-nodes
    #'(lambda (node)
	(if (= (tms::node-unique node) num) (return-from GET-NODE node)))))

(defun find-justification-with-antecedents (node nodes)
  (let ((losers nil))
    (dolist (j (tms::node-justifications node))
      (if (same-elements? (cdr j) nodes)
	  (push j losers)))
    losers))

(defun dump-justifications (&optional (stream *terminal-io*))
  (tms::walk-nodes
    #'(lambda (node)
	(format stream "~%~%For ~A, " (lprinc-string (view-node node)))
	(justifications node stream))))

(defun core-dump-justifications (out-file)
  (with-open-file (out out-file :direction :out)
    (dump-justifications out)))

;;;; Getting statistics on justifications

(proclaim '(special *j-table* *n-table*))

(defun classify-justifications ()
  (setq *n-table* nil)
  (tms::walk-nodes
    #'(lambda (n)
	(push (cons n (characterize-justifications (tms::node-justifications n)))
	      *n-table*)))
  (setq *j-table* (sort (invert-n-table *n-table*)
			#'(lambda (x y) (> (cdr x) (cdr y))))))

(defun characterize-justifications (j-list)
  (let ((table nil)
	(table-entry nil))
    (dolist (entry j-list)
      (setq table-entry (assoc (car entry) table :test #'equal))
      (cond (table-entry (incf (cdr table-entry)))
	    (t (push (cons (car entry) 1) table))))
    table))

(defun invert-n-table (in-table)
  (let ((ftable nil) (fentry nil))
    (dolist (entry in-table)
      (dolist (subentry (cdr entry))
	(setq fentry (assoc (car subentry) ftable :test #'equal))
	(cond (fentry (incf (cdr fentry) (cdr subentry)))
	      (t (push (cons (car subentry) (cdr subentry))
		       ftable)))))
    ftable))

(defun informant-distribution (&optional (stream *terminal-io*))
  ;; uses the J-table to produce a total count.
  (format stream "~%Justification Distribution by informant:~%")
  (do ((entries (cdr *j-table*) (cdr entries))
       (entry (car *j-table*) (car entries))
       (sum 0))
      ((null entry)
       (format stream "~%~30:<~>-----")
       (format stream "~%~30:<Total = ~>~6:<~D~>" sum))
    (format stream "~%~30:<~A~>~6:<~D~>" (car entry) (cdr entry))
    (incf sum (cdr entry))))

(defun show-bozo-justifications (&aux result)
  (tms::walk-nodes
    #'(lambda (node)
	(dolist (just (tms::node-justifications node))	
	  (when (member node (cdr just))
	    (format t "~% Justifying ~A via ~A using:~% ~A"
		    (lprinc-string (view-node node))
		    (car just)
		    (lprinc-string (cons 'and (mapcar #'view-node (cdr just)))))
	    (push (cons node just) result)))))
  result)

;;;; Finding nodes justified through particular informants

;; Handy for debugging -- get to see what your new code is really
;; asserting...

(defun nodes-justified-by (informant &aux result)
  (dolist (entry *n-table* result)
    (when (assoc informant (cdr entry))
      (push (car entry) result))))

;;;; Start-up tests

(defun shakedown (&optional (stream *standard-output*))
  ;; Runs test suite
  (format stream "~%Checking load state and packet system...")
  (test-packets-and-load stream)
  (format stream "~%Testing basic rule functionality...")
  (test-rules)
  (test-asns)
  (format stream "~%Testing rule nesting...")
  (test-nesting)
  (format stream "~%Performing N-queens tests...")
  (queens-test)
  (format stream "~%Basic ATMoSphere tests have finished successfully."))

(defun eval-run (forms &optional (stream *standard-output*) (silent? t))
  (dolist (form forms)
    (if silent? (eval form)
	(format stream "~A" (eval form)))
    (if silent? (run)
	(format stream "~A" (run)))))

(defun test-packets-and-load (&optional (stream *standard-output*) 
					&aux not-there too-much)
  ;; Sees if necessary rules are there
  (dolist (spec '((:test1 1) (:test-nest1 1)
		  (:test-nest2 1)(:test-nest3 1)(:test-nest4 1)
		  (:N-Queens-intern 1)(:N-Queens-in 1)
		  (:N-Queens-implied-by 1)))
    (let ((size (packet-defined? (car spec))))
      (cond ((integerp size)
	     (unless (= size (cadr spec))
	       (format stream "~% Packet ~A has ~D entries, should have ~D."
		       (car spec) size (cadr spec))
	       (setq too-much t)))
	    (t (format stream "~% Packet ~A not defined." (car spec))
	       (setq not-there t)))))
  (cond ((and not-there too-much)
	 (format stream
 "~% Please type (adb::clear-all-packets), load DEBUG and NQUEENS,~%and try again."))
	(not-there (format stream "~%Please load DEBUG and/or NQUEENS and try again."))))

;;;; Simplest tests

(store-in-packet :test1
  (rule :intern ((foo ?x) :var ?f1
		 (bar ?y) :var ?f2)
	(rjustify (Grumble ?x ?y) (?f1 ?f2) :Test1)))

(defun test-rules ()
  (init)
  (use-packet :test1)
  (eval-run '((assert! '(foo a))
	      (assert! '(bar b))
	      (assert! '(foo c))))
  (unless (fetch `((Grumble a b) . :TRUE))
    (error "Basic rule failure: test1"))
  (unless (fetch '((Grumble c b) . :TRUE))
    (error "Basic rule failure: ordering."))
  :OKAY)

(defun test-asns () ;; Truely paranoid
  (init)
  (use-packet :test1)
  (eval-run '((assume! '(foo a))
	      (assume! '(bar b))
	      (assume! '(foo c))))
  (unless (fetch `((Grumble a b) . :TRUE))
    (error "Basic rule failure on asn: test1"))
  (unless (fetch '((Grumble c b) . :TRUE))
    (error "Basic rule failure on asn: ordering."))
  :OKAY)

;;;; Testing rule nesting
(store-in-packet :test-nest1
   (rule :intern ((foo ?x) :var ?f1)
	 (rule :intern ((bar ?y) :var ?f2)
	       (rjustify (Intern-Intern ?x ?y) (?f1 ?f2)))))
(store-in-packet :test-nest2
   (rule :in ((foo ?x) :var ?f1)
	 (rule :in ((bar ?y) :var ?f2)
	       (rjustify (In-In ?x ?y) (?f1 ?f2)))))
(store-in-packet :test-nest3
   (rule :in ((foo ?x) :var ?f1)
	 (rule :intern ((bar ?y) :var ?f2)
	       (rjustify (In-Intern ?x ?y) (?f1 ?f2)))))
(store-in-packet :test-nest4
   (rule :intern ((foo ?x) :var ?f1)
	 (rule :in ((bar ?y) :var ?f2)
	       (rjustify (Intern-In ?x ?y) (?f1 ?f2)))))

(defun test-nesting ()
  (dolist (args '((:test-nest1 Intern-Intern)
		  (:test-nest2 In-In)
		  (:test-nest3 In-Intern)
		  (:test-nest4 Intern-In)))
    (apply #'test-in1 args)))

(defun test-in1 (pkt form)
  (init)
  (use-packet pkt) ;; Assuming at least one :IN trigger, and no :implied-by triggers
  (eval-run '((referent '((foo a) . :TRUE) t)
	      (referent '((bar b) . :TRUE) t)))
  (when (and (fetch `((,form a b) . :TRUE)) (not (eq form 'INTERN-INTERN)))
    (error "Nested rules ~A triggering prematurely" pkt))
  (eval-run '((assume! '(foo a))
	      (assume! '(bar b))))
  (unless (fetch `((,form a b) . :TRUE))
    (error "Nested rules ~A triggering incorrectly (asn-after case)." pkt))
  ':Okay)

;;;; Testing with N-queens

(defun queens-test (&aux solutions dds)
  (setq solutions (n-queens 4))
  (cond ((null solutions)
	 (error "Intern Queens failure: No solutions."))
	((> (length solutions) 2)
	 (error "Intern Queens failure: Too many solutions (~D instead of 2)."
		(length solutions)))
	((not (every #'(lambda (solution)
			 (okay-4-queens-solution? solution)) solutions))
	 (error "Bogus intern queens solutions: ~A"
		(remove-if-not #'okay-4-queens-solution? solutions))))
  (setq solutions (n-queens-in 4))
  (cond ((null solutions)
	 (error "In Queens failure: No solutions."))
	((> (length solutions) 2)
	 (error "In Queens failure: Too many solutions (~D instead of 2)."
		(length solutions)))
	((not (every #'(lambda (solution)
			 (okay-4-queens-solution? solution)) solutions))
	 (error "Bogus in queens solutions: ~A"
		(remove-if-not #'okay-4-queens-solution? solutions))))
  (setq dds (n-queens-inc 4))
  (setq solutions (try-next dds))
  (unless solutions
    (error "No solutions for implied-by queens."))
  (unless (okay-4-queens-solution? solutions)
    (error "Bogus implied-by queens solution: ~A" solutions))
  (setq solutions (try-next dds))
  (unless solutions
    (error " Too few solutions for implied-by queens."))
  (unless (okay-4-queens-solution? solutions)
    (error "Bogus implied-by queens solution: ~A" solutions))
  (setq solutions (try-next dds))
  (when solutions
    (error "Too many solutions for implied-by queens: ~A, (~A)"
	   solutions (if (okay-4-queens-solution? solutions) "okay" "bogus")))
  :okay)

(defun okay-4-queens-solution? (solution)
  (some #'(lambda (answer)
	    (every #'(lambda (fact)
		       (implied-by? fact solution)) answer))
	'((((QUEEN 0) . 1) ((QUEEN 1) . 3)
	   ((QUEEN 2) . 0) ((QUEEN 3) . 2))
	  (((QUEEN 0) . 2) ((QUEEN 1) . 0)
	   ((QUEEN 2) . 3) ((QUEEN 3) . 1)))))



;;;; Compactly printing antecedent graph

;; ******** Here at least are the design notes.  Later.

;;This is text-oriented both for portability and because graphical systems tend to
;; waste screen real-estate.

;; Organization: First walk backward breadth-first, gathering up alist of nodes and depth.
;;  (Want to print repeated nodes at minimal depth since indenting will be smaller)
;;  Include different criteria, such as depth bound and hitting assumptions.
;;  Mark paths which are loops by a prefix of some kind.


;;;; Checking for datastructure integrity violations

;;; One should never have two asserts which have eq
;;; lisp forms.

(defun check-assert-integrity ()
  (setq *assert-table* nil *losers* nil)
  (maphash #'(lambda (key entry)
	       (let ((e (assoc (lisp-form entry) *assert-table*)))
		 (if e (push entry (cdr e))
		     (push (list (lisp-form entry) entry) *assert-table*))))
	   *fact-hash-table*)
  (dolist (entry *assert-table* (progn (setq *assert-table* nil) *losers*))
    (when (cddr entry)
      (push entry *losers*)
      (format t "~% Loser: ~A" entry))))

(defun check-justification-e2-propagation (&aux losers)
  ;; Checks to see if antecedents have empty env (e2) but consequent
  ;; has something else.  Inspired by bug of 1/15/90, v21 ATMS
  (dolist (node tms::*nodes* losers)
    (unless (and (tms::node-envs node)
		 (eq (car (tms::node-envs node)) tms::*empty-env*))
      (dolist (just (tms::node-justifications node))
	(when (every #'(lambda (node)
			 (eq (car (tms::node-envs node)) tms::*empty-env*))
		     (cdr just))
	  (pushnew node losers)
	  (format t "~% Found a losing node: ~A" (view-node node))
	  (return))))))
