;; -*- Mode: lisp; Syntax: Common-lisp; Package: ADB; -*-

(in-package 'ADB)

;;****** Need to put timers back 

;;;; Utilities for writing interpretation constructors
;;
;;Some subset of the data will be premises.  Usually constructing
;;global interpretations (i.e., using every premise in the TMS) is undesirable
;;because it might be representing several completely independent problems.
;;Part of work involved in writing interpretation constructors is gathering up
;;the necessary facts.  To begin with, here are some utilities for that.

(export '(try-extension find-extensions
	  show-interps find-extensions2 find-all-consistent-combinations
	  careful-ordered-insert find-all-consistent-maximal-combinations
	  whyq why-notq why justifications
	  why-not consequences possible-values? pvs compatible-facts?
	  fact-consistent-in? implied-by? in? out? environments
	  fact-assumption create-environment environment-cons
	  lookup-fact find-fact-union-killer
	  envs-common-env envs-common-assumptions
	  env-residue-env env-residue find-subsuming-contradiction
	  assumption-alternates get-assumption-extensions nogoods-including))

(defun try-extension (seed-env fact &aux node env)
  ;assumes fact is an assumption
  (setq node (get-tms-node fact)
	env (tms::cons-env seed-env (tms::node-assumption node)))
  (unless (tms::env-contradictory env) env))

(defun try-node-extension (seed-env node &aux env)
  ;assumes node has an assumption
  (setq env (tms::cons-env seed-env (tms::node-assumption node)))
  (unless (tms::env-contradictory env) env))

;;;; My own interpretation constructor

(defun find-extensions (seed-env facts)
;;;Assumes all facts are already assumptions 
;;;Now prunes out true and false alternatives.
  (do ((alternatives facts (cdr alternatives))
       (interpretations (list seed-env))
       (previous-interpretation nil))
      ((null alternatives) interpretations)
    (if-debug-flag extensions
		   (format t "~%Starting new alternative set:")
		   (pprint (car alternatives)))
    (do ((alternative (car alternatives) (cdr alternative))
	 (node nil)
	 (new nil)
	 (new-interps nil))
	((null alternative) (setq previous-interpretation interpretations
				  interpretations new-interps)
	 (if-debug-flag count-extensions
			(format t "~%    Now ~A interpretations."
				(length new-interps))))
      (if-debug-flag extensions
		     (format t "~%      Trying ~A" (car alternative)))
      (setq node (get-tms-node (car alternative)))
      (when (tms::true? node) (setq alternative nil))
      (unless (tms::false? node) 
	(dolist (i interpretations)
	(setq new (try-node-extension i node))
	(if new (push new new-interps))
	(if-debug-flag extensions
		       (cond (new (format t "~%New extension:~%~A"
					  (print-env new)))
			     (t (format t "~%Contradictory."))))
	(if-debug-flag extensions-break
		       (break 'new-extension t)))))))

(defun show-interps (interps &optional (pause? nil))
  (dolist (i interps)
    (print-env i)
    (if pause? (read-char))))

;;;; Experimental version with incremental pruning

(defun find-extensions2 (seed-env facts &optional (pruner nil)
			 &aux interpretations previous subsumed?
			 new-interps as alt-ass
			 new)
;;; Assumes all facts are already assumptions
  (setq interpretations (list seed-env)
	previous nil)
  (dolist (alternatives facts) ;Gack -- Zetalisp dolist won't take return arg!
    ;;;For each env in the interpretations, see if it already implies one
    ;;;  of the alternatives provided
    (setq new-interps nil)
    (dolist (i interpretations)
      (setq subsumed? nil alt-ass nil)
      (dolist (alternative alternatives)
	(setq as (get-tms-node alternative))
	(cond ((tms::true-in? as i) ;implied by the interp already
	       (setq subsumed? as)
	       (return as))
	       ;; o.w., build up lists of assumptions
	      (t (push (tms::node-assumption as) alt-ass))))
      (cond (subsumed?
	     (push (tms::cons-env i (tms::node-assumption as)) new-interps))
	    (t ;;;; Try each one and see if it works.
	(dolist (alt-as alt-ass)
	  (setq new (tms::cons-env i alt-as))
	  (unless (tms::env-contradictory new)
	    (push new new-interps))))))
      ;;; Increment the other parameters
    (setq previous interpretations
	  interpretations new-interps)
	 (if-debug-flag count-extensions
			(format t "~%    Now ~D interpretations."
				(length new-interps)))
	 (when (and pruner
		    (> (length interpretations)
		       (/ (* (length alternatives)
			      (length previous)) 2.0)))
	   (setq interpretations
		 (remove-if #'(lambda (e) (tms::env-contradictory e))
			 (remove-if #'(lambda (e)
				     (or (tms::env-contradictory e)
					 (funcall pruner e))) interpretations)))
	   (if-debug-flag count-extensions
		 (format t "~%       After pruning, ~D"
			 (length interpretations)))))
  interpretations)

;;;; Faster version, for internal use only.
;;    Takes choice sets specified by lists of assumptions

(defun find-assumption-extensions (seed-env choice-sets &aux current-solutions)
  (if-debug-flag count-extensions
    (let ((lengths (mapcar #'length choice-sets)))
      (format t "~%  ~A choice sets, lengths~%   ~A~% Worst case = ~A solutions."
	      (length choice-sets) lengths
	      (apply #'* lengths))))
  (setq current-solutions
	(tms::variable-interpretations-depth-stack-sets seed-env choice-sets))
  current-solutions)

;; Alot of the other tests, like membership & contradictory, are already done in cons-env
;; more efficiently than we could.  But ordering by size is our department.  Instead of
;; doing a total sort, just split according to lengths 1, 2 and many.

(defun revise-asn-choice-sets (choice-sets &aux ones twos manys)
  ;; Re-orders them by size, to minimize branching.
  (dolist (choice-set choice-sets (nconc ones twos manys))
    (cond ((null choice-set) (return-from revise-asn-choice-sets nil))
	  ((cddr choice-set) (push choice-set manys))
	  ((cdr choice-set) (push choice-set twos))
	  (t (push choice-set ones)))))

;;;; putting together consistent collections of facts
;;
;; This code is a simple way of generating the subset of consistent sets
;; from the power set of a collection of facts. 
;; It assumes the facts are totally independent.
;; Not clear if this is really useful, however, see below.

(defun find-all-consistent-combinations (facts)
  (let ((fe-pairs (mapcar #'(lambda (fact)
			      (cons (list fact)
				    (tms::label (get-tms-node fact))))
			  facts)))
    (append fe-pairs (find-all-consistent-combinations1 fe-pairs fe-pairs))))

(defun find-all-consistent-combinations1 (so-far extension-candidates)
  (do ((basics (cdr so-far) (cdr basics))
       (basic (car so-far) (car basics))
       (extensions nil))
      ((null basic)
       (if extensions
	   (append extensions (find-all-consistent-combinations1
				extensions
				extension-candidates))))
    (do ((candidate (car extension-candidates) (car candidates))
	 (candidates (cdr extension-candidates) (cdr candidates))
	 (envs nil)
	 (nfacts nil))
	((null candidate))
      (setq nfacts (careful-ordered-insert (caar candidate) (car basic)))
      (cond ((null nfacts)) ;skip if already included
	    ((some extensions ;ignore if already found
		   #'(lambda (fe-pair) (equal (car fe-pair) nfacts))))
	    ((setq envs (tms::union-envss (cdr candidate) (cdr basic)))
	     ;if not inconsistent, then add to extensions
	     (push (cons nfacts envs) extensions))))))

(defun careful-ordered-insert (new-thing ll)
  (catch 'careful-ordered-insert1 
	 (careful-ordered-insert1 new-thing ll)))

(defun careful-ordered-insert1 (new-thing ll)
  ;returns nil if new-thing is already in the list
  ;;**** GET RID OF ALPHALESSP
  ;;**** CHANGE TO SHARE STRUCTURE
  (cond ((null ll) (cons new-thing nil))
	((equal new-thing (car ll)) (throw 'careful-ordered-insert1 nil))
	((alphalessp (car ll) new-thing)
	 (cons (car ll) (careful-ordered-insert1 new-thing (cdr ll))))
	(t (cons new-thing ll))))

;; Making closed-world assumptions in a consequent fashion requires finding all
;; MAXIMAL subsets of a set of facts which can be taken together.  This code does it.
;; It assumes as input a list of alternative sets (as per the extensions code)
;; but will only return those sets which include one choice from each alternative.

(defun find-all-consistent-maximal-combinations (fact-sets &aux results)
  (dolist (alt (car fact-sets))
    (if (in? alt) ;everything is relative
	(setq results (nconc (find-all-consistent-maximal-combinations1
				(cons (cons alt nil)
				      (environments alt)) ;pass down envs
				(cdr fact-sets))
			      results))))
  results)

(defun find-all-consistent-maximal-combinations1 (seed rest &aux nenvs result)
  (cond ((null rest) (cons (car seed) nil)) ;got one
	(t (dolist (alt (car rest))
	     (when (setq nenvs (tms::union-envss (cdr seed)
					      (environments alt)))
	       ;consistent extension, so try for maximal
	       (setq result
		     (nconc (find-all-consistent-maximal-combinations1
			       (cons (cons alt (car seed)) nenvs) (cdr rest))
			     result))))
	   result)))

;;;; Finding extensions by adding classes of assumptions
;;
;; A common problem, at least for QPE, is extending an env by a particular collection
;; of assumptions.  This program will find them.

(defun get-assumption-extensions (base-env assumption-sets
				  &aux solutions new-solution old-solutions)
  (setq old-solutions (list base-env))
  (dolist (assumption-set assumption-sets)
    (dolist (assumption assumption-set)
      (dolist (old-env old-solutions)
	(setq new-solution (tms::cons-env old-env assumption))
	(unless (tms::env-contradictory new-solution)
	  (push new-solution solutions))))
    (setq old-solutions solutions
	  solutions nil))
  old-solutions)

;;; Some information is needed to order classes in interpretation construction.
;;; One possibility is to use those elements which are most constrained first.
;;; The function below attempts to test this.

(defun node-constraints (node)
  (+ (length (tms::node-consequents node))
     (length (tms::node-classes node))))

