;;; -*- Mode: LISP; Syntax: Common-Lisp; Package: ADB; -*-

(in-package 'adb)

;;; Programmable Common-Lisp Logic-oriented  pretty printer
;;

;; Private copy for ADB.

;; Overall design: Given a form, the FORM-PARSER turns it into a list of TOKENS.
;; Each token is either a string or a marker.  Markers include :NEWLINE, :START,
;; and :END.  :START and :END delimit blocks which should be treated as a unit.
;; :LP and :RP indicate parentheses.  :STARTL, :ENDL indicate blocks which should
;; be indented "lisp-style".
;;
;; The FORM-PRINTER is responsible for taking a parsed form and printing it in
;; a reasonable fashion on a given stream.  There need to be different entrances
;; into the function in order to capture different purposes.
;; Additional parameters,
;; which for some versions can be extracted directly from the stream, include
;; *LINE-LENGTH*, *INITIAL-INDENTATION*.  The latter helps for using the printer in
;; report generation.

;; ***** Don't really need parens everywhere.  The printer could parse the form
;; ***** decide where to put them in based on :START-:START combinations?

(export '(lprint lprinc-string defForm-Parser defForm-Parser-cadr
		 parse-form parse-argument-list))

;; Parameters of stream
(defvar *LINE-LENGTH* 80.) ;; Default length of line
(defvar *INITIAL-INDENTATION* 0) ;; Default indentation

;; Internal variables
(defvar *PAREN-FLAG* T) ;; If NIL, use "()".  If T, use "[]".
(defvar *report-mode* :ASCII)  ;; Controls how things are printed
;; Supported modes include
;;  :ASCII -- absolutely vanilla, should come out okay on anything
;;  :GRAPHICS -- for showing stuff on a screen with fancy characters.
;;               it is assumed sharp-sign switches will be used to rebind
;;               the symbols for the particular machine being used.
;;  :LaTeX -- for producing LaTeX reports

;; Useful predicates

(defun start? (thing) (or (eq thing :START) (eq thing :STARTL)))

(defun end? (thing) (or (eq thing :END) (eq thing :ENDL)))

(defun form-function? (thing)
  (cond ((null thing) nil)
	((symbolp thing) (get thing 'TYPE-OF-FUNCTION?))
	((not (listp thing)) nil)
	(t (form-function? (car thing)))))

(defun get-form-parser (thing)
  (cond ((null thing) nil)
	((symbolp thing) (get thing 'form-parser))
	(t nil)))

(defun get-form-cadr-parser (thing)
  (cond ((null thing) nil)
	((symbolp thing) (get thing 'cadr-form-parser))
	(t nil)))

(defmacro DefForm-Function (fun) `(setf (get ',fun 'TYPE-OF-FUNCTION?) t))

;; Simplistic top-level

(defun lprint (form &optional (stream *terminal-io*))
  (terpri stream)
  (form-printer (strip-initial-parens (parse-form form)) stream))

(defun lprinc (form &optional (stream *terminal-io*))
  (form-printer (strip-initial-parens (parse-form form)) stream))

(defun lprinta (forms &optional (stream *terminal-io*))
  (dolist (form forms) (lprint form stream)))

;;;; First-pass version of FORM-PARSER
;;
;; For programmability, make it data-driven.

(defun parse-form (form &aux parser)
  (cond ((null form) nil)
	((not (listp form)) (stringify form))
	((setq parser (get-form-parser (car form)))
	 ;; Use data-driven function.
	 (funcall parser form))
	((not (listp (cdr form))) ;; means a function.
	 ;; In general version, move this code into some
	 ;; programmable function.  This version is ADB-specific
	 ;; and thus we'll optimize.
	 (case (cdr form)
	   (:TRUE (parse-form (car form)))
	   (:FALSE (parse-form (list ':NOT (car form))))
	   (t (list :START :LP (parse-form (car form))
		       "=" (parse-form (cdr form)) :RP :END))))
	((form-function? (car form)) ;; If function, produce equation form.
	 (list :START :LP (parse-form (car form))
	       "=" (parse-form (cdr form)) :RP :END))
	((setq parser (get-form-cadr-parser (cadr form))) ;; check out internal form
	 (funcall parser form))
	(t (nconc (list :START (stringify (car form))
		    "(" )
		  (parse-argument-list (cdr form))
		  (list ")" :END)))))

(defun stringify (x) ;; Should add something to this for flushing package names!
  (format nil "~A" x))

(defun parse-argument-list (args &optional (spacer-string ","))
  (cond ((null args) nil)
	((not (listp args)) (parse-form args))
	((cdr args)
	 (cons (parse-form (car args))
	       (cons spacer-string (parse-argument-list (cdr args)
							spacer-string))))
	(t (list (parse-form (car args))))))

;;;; Defining form parsers.
;; Use conventions for open-coded unification to simplify things.
;; Other systems can now produce assertions which can tie into the
;; pretty-printer.

(defun var? (x) (and (symbolp x) (string= "?" (symbol-name x)
					   :end2 1))) ;; Get this from other stuff.

(defun produce-binding-list (trigger)
;;Produces an alist of (<var> . <path>) which can be used to 
;; extract the trigger variables.
 (produce-binding-list1 trigger 'form nil))
;; For simplicity, assume the form matches.

(defun produce-binding-list1 (trigger path so-far)
  (cond ((null trigger) so-far)
	((var? trigger)
	 (cons (list trigger path) so-far))
	((not (listp trigger)) so-far)
	(t (produce-binding-list1 (cdr trigger) (list 'CDR path)
	    (produce-binding-list1 (car trigger) (list 'CAR path) so-far)))))

(defmacro DefForm-Parser (trigger &rest body)
  `(setf (get ',(car trigger) 'FORM-PARSER)
	 #'(lambda (form) (let ,(produce-binding-list trigger)
			    ,@ body))))

(defmacro DefForm-Parser-cadr (trigger &rest body)
  `(setf (get ',(cadr trigger) 'CADR-FORM-PARSER)
	 #'(lambda (form) (let ,(produce-binding-list trigger)
			    ,@ body))))

;;;; First-pass form printer
;;
;; Ignore all the niceties, just spit out the stuff recursively.

(defun form-printer (form &optional (stream *terminal-io*)) ;takes a parsed form
  (cond ((null form))
	((listp form)
	 (cond ((symbolp (car form))
		(case (car form)
		  (:start (form-printer (cdr form) stream))
		  (:end (form-printer (cdr form) stream))
		  (:LP (print-left-paren stream)
		       (let ((*paren-flag* (not *paren-flag*)))
			 (form-printer (cdr form) stream)))
		  (:RP (print-right-paren stream)
		       (form-printer (cdr form) stream))
		  (t (error "Shouldn't be a symbol here -- ~A" form))))
	       ((listp form) (form-printer (car form) stream)
		(form-printer (cdr form) stream))
	       (t (format stream "~A" (car form))
		  (form-printer (cdr form) stream))))
	(t (format stream "~A" form))))

(defun print-left-paren (&optional (stream *terminal-io*))
  (cond (*paren-flag* (format stream "("))
        (t (format stream "["))))

(defun print-right-paren (&optional (stream *terminal-io*))
  (cond (*paren-flag* (format stream "]"))
	(t (format stream ")"))))

(defun strip-initial-parens (parsed-form)
  ;; gets rid of top-level parentheses.
  (cond ((not (listp parsed-form)) parsed-form)
	((and (eq (car parsed-form) ':START)
	     (eq (cadr parsed-form) ':LP)) ;; Assume it ends well
	 (setq parsed-form (nreverse parsed-form))
	 (setf (cdr parsed-form) (cddr parsed-form))
	 (setq parsed-form (nreverse parsed-form))
	 (setf (cdr parsed-form) (cddr parsed-form))
	 (strip-internal-negation-parens parsed-form))
	(t (strip-internal-negation-parens parsed-form))))

(defun strip-internal-negation-parens (parsed-form)
  (cond ((null parsed-form) nil)
	((not (listp parsed-form)) parsed-form)
	(t (mapcar #'(lambda (pform)
		       (cond ((not (listp pform)) pform)
			     ((and (eq (car pform) :START)
				   (eq (cadr pform) :LP)
				   (eq (third pform) (get-not-symbol)))
			      `(:START ,(get-not-symbol)
				       ,(strip-internal-negation-parens (fourth pform))
				       :END))
			     (t (strip-internal-negation-parens pform))))
		   parsed-form))))

;;;;; Building fancy strings

(defun lprinc-string (form)
  (form-stringer (strip-initial-parens (parse-form form))))

(defun form-stringer (form)
  (cond ((null form) "")
	((listp form)
	 (cond ((symbolp (car form))
		(case (car form)
		  (:start (form-stringer (cdr form)))
		  (:end (form-stringer (cdr form)))
		  (:LP (concatenate 'string (print-left-paren nil)
				      (let ((*paren-flag* (not *paren-flag*)))
					(form-stringer (cdr form)))))
		  (:RP (concatenate 'string (print-right-paren nil)
				      (form-stringer (cdr form))))
		  (t (error "Shouldn't be a symbol here -- ~A" form))))
	       ((listp form) (concatenate 'string (form-stringer (car form))
					    (form-stringer (cdr form))))
	       (t (concatenate 'string (format nil "~A" (car form))
				 (form-stringer (cdr form))))))
	(t (format nil "~A" form))))			  

;;;; Some special symbols

(defvar *AND-SYMBOL* #+Symbolics ""
	#-Symbolics "&")
(defvar *OR-SYMBOL* #+Symbolics ""
	#-Symbolics "|")
(defvar *NOT-SYMBOL* #+Symbolics ""
	#-Symbolics "~")
(defvar *IMP-SYMBOL* #+Symbolics ""
	#-Symbolics "->")
(defvar *EQV-SYMBOL* #+Symbolics ""
	#-Symbolics "<->")
(defvar *LEQ-SYMBOL* #+Symbolics ""
	#-Symbolics "<=")
(defvar *GEQ-SYMBOL* #+Symbolics ""
	#-Symbolics ">=")
(defvar *UNREL-SYMBOL* #+Symbolics ""
	#-Symbolics "??"
	)

(defun get-and-symbol ()
  (case *report-mode*
    (:ASCII  "&")
    (:GRAPHICS *and-symbol*)
    (:LATEX "\\( \\wedge \\)")))

(defun get-or-symbol ()
  (case *report-mode*
    (:ASCII  "|")
    (:GRAPHICS *or-symbol*)
    (:LATEX "\\( \\vee \\)")))

(defun get-not-symbol ()
  (case *report-mode*
    (:ASCII  "~")
    (:GRAPHICS *not-symbol*)
    (:LATEX "\\( \\neg \\)")))

(defun get-imp-symbol ()
  (case *report-mode*
    (:ASCII  "->")
    (:GRAPHICS *imp-symbol*)
    (:LATEX "\\( \\Rightarrow \\)")))

(defun get-eqv-symbol ()
  (case *report-mode*
    (:ASCII  "<->")
    (:GRAPHICS *eqv-symbol*)
    (:LATEX "\\( \\Leftrightarrow \\)")))

(defun get-lt-symbol ()
  (case *report-mode*
    ((:ASCII :GRAPHICS)  "<")
    (:LATEX " \\( < \\) ")))

(defun get-gt-symbol ()
  (case *report-mode*
    ((:ASCII :GRAPHICS)  ">")
    (:LATEX " \\( > \\) ")))

(defun get-eq-symbol ()
  (case *report-mode*
    ((:ASCII :GRAPHICS)  "=")
    (:LATEX " \\( = \\) ")))

(defun get-leq-symbol ()
  (case *report-mode*
    (:ASCII  "<=")
    (:GRAPHICS *leq-symbol*)
    (:LATEX "\\( \\leq \\)")))

(defun get-geq-symbol ()
  (case *report-mode*
    (:ASCII  ">=")
    (:GRAPHICS *geq-symbol*)
    (:LATEX "\\( \\geq \\)")))

(defun get-unrelated-symbol ()
  (case *report-mode*
    (:ASCII  "??")
    (:GRAPHICS *unrel-symbol*)
    (:LATEX "\\( \\downarrow \\)")))
