;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: ADB; -*-

(in-package 'adb)

;;; N-queens examples

;;; Three different versions are included:
;;
;; N-QUEENS uses intern rules, has optional flag to switch
;; between depth-first and breadth-first interpretation construction.
;;
;; N-QUEENS-IN uses in rules, the standard consumer architecture.
;;
;; N-QUEENS-INC uses focus environments and implied-by rules to
;; generate solutions incrementally.  It returns a generator
;; (also bound to the global variable *dds*) which can be "poked"
;; whenever a new solution is needed by calling (TRY-NEXT <generator>).
;; TRY-NEXT returns either the next solution environment or NIL to
;; indicate exhaustion.

(defun n-queens (n &aux csets choices)
  ;; Uses intern rules
  (adb:init t)
  (use-packet :N-Queens-Intern)
  (setq csets nil)
  (dotimes (x n)
    (setq choices nil)
    (dotimes (y n)
      (let ((choice `((Queen ,x) . ,y)))
	(push choice choices)
	(assume! choice)))
    (push choices csets))
  (run)
  (setq *solutions* (find-extensions tms::*empty-env*
				     csets)))

(defun n-queens-in (n &aux csets choices)
  ;; Uses intern rules
  (adb:init t)
  (use-packet :N-Queens-In)
  (setq csets nil)
  (dotimes (x n)
    (setq choices nil)
    (dotimes (y n)
      (let ((choice `((Queen ,x) . ,y)))
	(push choice choices)
	(assume! choice)))
    (push choices csets))
  (run)
  (setq *solutions* (find-extensions tms::*empty-env*
				     csets)))

(defun queens-okay? (x1 y1 x2 y2)
  (if (= y1 y2) nil
      (if (= (abs (- x1 x2)) (abs (- y1 y2))) nil
	  t)))

(defun test-queens (from to)
  (do ((n from (1+ n)))
      ((> n to))
    (time (n-queens n)) ;; from my init file
    (format t "~%For ~D queens, ~D solutions." n (length *solutions*))))

;;;; Implied-by version.
;;
;; Tuned to produce solutions incrementally
;; Slightly more complicated.

(defvar *dds* nil)

(proclaim '(special *dds*))

(defun n-queens-inc (n)
  (init t)
  (use-packet :N-Queens-Implied-by)
  (create-queens-dds-state n)
  ;(try-next *dds*) ;; returns one solution.
  )

(defun create-queens-dds-state (n)
  (setq *dds* nil)
  (dotimes (x n)
    (let ((current-choice-set nil))
      (dotimes (y n)
	(let ((choice `((Queen ,x) . ,y)))
	  (push choice current-choice-set)
	  (assume! choice)))
      (push (make-choice-set :choices  current-choice-set
			     :queue current-choice-set) *dds*)))
  (setq *dds* (make-dds-state :status ':INITIAL
				      :choice-sets *dds*)))

;;;; Packets for each version

(Store-in-Packet :N-Queens-Intern
 (rule :INTERN (((Queen ?x1) . ?y1) :var ?f1
	       ((Queen ?x2) . ?y2) :var ?f2
	       :test (not (or (= ?x1 ?x2)
			      (Queens-okay? ?x1 ?y1 ?x2 ?y2))))
      (rnogood (?f1 ?f2))))

(Store-In-Packet :N-Queens-In
 (rule :IN (((Queen ?x1) . ?y1) :var ?f1
	    ((Queen ?x2) . ?y2) :var ?f2
	    :test (not (or (= ?x1 ?x2)
			   (Queens-okay? ?x1 ?y1 ?x2 ?y2))))
       (rnogood (?f1 ?f2))))

(Store-In-Packet :N-Queens-Implied-By
 (rule :implied-by (((Queen ?x1) . ?y1) :var ?f1
		    ((Queen ?x2) . ?y2) :var ?f2
		    :test (not (or (= ?x1 ?x2)
				   (Queens-okay? ?x1 ?y1 ?x2 ?y2))))
       (rnogood (?f1 ?f2))))