;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: ADB; -*-

(in-package 'adb)

;;;; ATMoSphere rule system


;; Triggering conditions supported:
;;  :INTERN          -- runs when all triggers are added to the database,
;;                      independent of belief status.
;;  :IN              -- runs when the union of the trigger nodes have a
;;                      non-empty label.
;;  :IMPLIED-BY      -- runs when every trigger node is implied by the env bound
;;                      to the variable *FOCUS*.

(proclaim '(special *IN-NODES* *IMP-NODES*
		    *bound-vars* *end-forms*
		    *intern-queue* *in-queue*
		    *implied-by-queue* *implied-by-checked*
		    *packets* *default-packet*
		    *rule-name-counter* *rule-form-counter*
		    *rule-file-identifier*))

(export '(rule rnogood rassert! rassume refer-to rjustify quotize nlet clear-queues
	  enqueue dequeue queue-empty? run rules-file))

(defvar *BOUND-VARS* NIL) ;this is used by the rule code to keep track of
 ;the variables bound so far while expanding rules

(defvar *noisy?* nil) ;turns on announcement of matching in rules. 
;; Useful when trying to figure out why a particular rule doesn't work.

(defvar *rule-file-identifier* (symbol-name (gensym)))
(defvar *rule-name-counter* 0) ;uniquizer for rule names
(defvar *rule-form-counter* 0)

(defmacro rules-file (id-string)
  (setq *rule-file-identifier* id-string)
  (setq *rule-name-counter* 0) nil)

(defvar *intern-queue* nil) ;; queue for intern rules
(defvar *in-queue* nil) ;; queue for in rules
(defvar *implied-by-queue* nil) ;; queue for implied-by rules
(defvar *implied-by-checked* nil)

(defvar *focus* nil) ;; Environment which constitutes the center of attention. 

;;;; Rule syntax

;; (rule <condition> <triggers> . <body>)
;;     <condition> = :IN or :INTERN or :IMPLIED-BY.
;;     <triggers> = list of <form> | <form> :var <variable> |
;;                    | <form> :test <code> | combinations of them
;;     <body> = arbitrary lisp code

;;;; Design:

;; :INTERN rules are straightforward.
;; :IN rules should expand into normal ATMS consumers.  This part is broken right now.
;;     Instead, explicit unioning of the labels for the :IN triggers occurs.
;; :IMPLIED-BY rules are executed whenever their triggers are all implied-by the
;;     focus envirnoment.

;; Triggers can now be mixed arbitrarily.  All nested :IN triggers are interpreted
;; conjunctively, that is, the union of all such nodes must have a non-empty common
;; label before triggering will occur.  Furthermore, :IMPLIED-BY triggers take priority
;; over :IN triggers.  That is, if there are both kinds of triggers, the :IMPLIED-BY
;; triggers must be satisfied before the :IN triggers will even be examined.
;; This supports the control regieme described in the AAAI-88 paper.

;; Queues.

;;;; Building and installing rules

;; We do as much work as possible at rule expansion time to allow
;; the lisp compiler to do as much optimization as possible.

(proclaim '(special *end-forms* *condition*))

(defvar *desugar-triggers?* nil)

(defmacro srule (condition trigger &rest body)
  (let ((*desugar-triggers?* t))
    (macroexpand `(rule ,condition ,trigger ,@ body))))

(defmacro rule (condition triggers &rest body)
  (let ((*end-forms* nil)
	(parsed-triggers (parse-nested-triggers triggers))
	(*condition* condition)) ;; make condition accessible.
      (add-rule (car parsed-triggers)
		(subst 'internal-rule 'rule
		       (make-nested-rule (cdr parsed-triggers) body))
		condition)))

(defmacro internal-rule (condition triggers &rest body)
  (let ((*condition* condition)
	(parsed-triggers (parse-nested-triggers triggers)))
    `(add-internal-rule ,(car parsed-triggers)
			,(make-nested-rule (cdr parsed-triggers) body)
			,condition)))

(defun make-nested-rule (triggers body)
  (cond ((null triggers) body)
	(t `((add-internal-rule ,(car triggers)
			    ,(make-nested-rule (cdr triggers) body)
			    ,*condition*)))))

(defun add-rule (trigger body condition)
  (let ((top-indexing (subst nil '*IN-NODES*
			     (subst nil '*IMP-NODES* (build-rule trigger body condition))))) 
    `(progn ,@ *end-forms*
	    (cond ((okay-packet? *default-packet*)
		   (add-to-packet *default-packet* ',top-indexing))
		  (t ,top-indexing)))))

(defmacro add-internal-rule (trigger body condition) (build-rule trigger body condition))

;;;; Parsing triggers
;;
;; Output is a list of the form (<pattern> <form-var> <test>)

(defun parse-nested-triggers (triggers)
  (labels ((parse-trigger (triggers &optional (var nil) (test nil))
	    (case (car triggers)
		  (:var (parse-trigger (cddr triggers) (cadr triggers) test))
		  (:test (parse-trigger (cddr triggers) var (cadr triggers)))
		  (t (values var test triggers)))))
	 (cond ((null triggers) nil)
	       (t (multiple-value-bind (var test new-triggers)
		      (parse-trigger (cdr triggers))
		    (cons (list (car triggers) var test)
			  (parse-nested-triggers new-triggers)))))))

;;;; Building a rule.
;;    involves analyzing the trigger to see what
;;   special-purpose unifier is needed, constructing a function
;;   to do the work of the body, and building a form to index it.

(defvar *Sweeten-trigger-patterns?* t) ;; Allow syntactic sugar in patterns

(defun build-rule (trigger body condition &aux match-function
			   body-function)
  (let ((pattern (desugar-pattern (car trigger)))
	(var (cadr trigger))
	(test (caddr trigger)))
  (setq match-function (generate-match-function pattern var test condition))
  (setq body-function (generate-body-function pattern var body))
  (push match-function *end-forms*)
  (push body-function *end-forms*)
  `(insert-rule (lookup-class ,(get-trigger-class pattern) t)
		(function
		  ,(if *bound-vars*
		      `(lambda (p)
			 (,(cadr match-function) p ,@ *bound-vars*))
		       (cadr match-function)))
		(function
		  ,(if *bound-vars*
		       (let ((tv (nreverse (pattern-free-variables trigger))))
			 `(lambda ,tv
			  (,(cadr body-function) ,@ tv
			   ,@ (scratchout tv *bound-vars*))))
		       (cadr body-function)))
		*IN-NODES*
		*IMP-NODES*)))

(defun get-trigger-class (trigger)
  (cond ((and (not (listp trigger))
	      (not (member trigger *bound-vars*))) ;barf
	 (error "~%Trigger must be a list -- ~A." trigger))
	(t (find-trigger-class trigger))))

(defun find-trigger-class (trigger)
  (cond ((variable? trigger)
	 (if (member trigger *bound-vars* :test #'eq) 
	     `(get-assertion-class ,trigger) ;have to test it at run-time
	     (error "~%Trigger class is unbound -- ~A." trigger)))
	((not (listp trigger)) (list 'QUOTE trigger))
	(t (find-trigger-class (car trigger)))))

;;;; Desugaring patterns

;;; Nobody likes to type (foo . :TRUE).  So we allow them to type other things

(defun desugar-pattern (pattern)
  (cond ((not *Sweeten-trigger-patterns?*) pattern)
	((not (listp pattern)) pattern)
	((not (listp (cdr pattern)))
	 ;; Assumes old-style explicit declaration
	 pattern)
	(t (case (car pattern) ;; Install special stuff
	     (:EVAL pattern) ;; Return itself
	     (NOT (cons (cadr pattern) :FALSE))
	     ;; Assumes specification of function (I don't want to take "=",
	     ;; unfair to mathematics hackers).
	     (:== (cons (cadr pattern) (caddr pattern)))
	     ;; Assume normal predication as default.
	     (t (cons pattern :TRUE))))))

(defun desugar-all-pattern (pattern)
  ;; recursive descent
  (cond ((null pattern) nil)
	((listp pattern)
	 (case (car pattern)
	   (NOT (let ((internal (desugar-all-pattern (cadr pattern))))
		  (case (cdr internal)
		    (:TRUE (cons (car internal) :FALSE))
		    (:FALSE (cons (car internal) :TRUE))
		    (t (error "~% ~A should be true or false (CDR ~A)."
			      (cdr internal) internal)))))
	   (t (desugar-pattern
	   (cons (car pattern) ;; assume no implicit sets
		 (mapcar #'desugar-all-pattern (cdr pattern)))))))
	(t (cons pattern :TRUE))))

;;;; Generating the body function
;;
;; This is radically simpler since the condition checking has been moved to
;; the run-time system. 

(defun generate-body-function (pattern var body
				       &aux fname newly-bound env)
  (setq newly-bound (if var (cons var (pattern-free-variables pattern))
			(pattern-free-variables pattern)))
  (setq body (with-pushed-variable-bindings newly-bound (fully-expand-body body)))
  (setq env (append newly-bound (scratchout newly-bound *bound-vars*)))
  (setq fname (generate-rule-function-name pattern)) 
  `(defun ,fname ,env
     ,@ (if *noisy?* `((progn (format t "~%Bound ~A."
				      ,(quotize pattern))
			      ,@body))
	    body)))
     
(defun generate-rule-function-name (pattern)
 (intern (format nil "~A-~A-~A" pattern *rule-file-identifier*
		 (incf *rule-name-counter*))))

(defun generate-variable (prefix)
  (intern (format nil "?~A~D" prefix (incf *rule-form-counter*))))

;;;; Open-coding unification

(defun generate-match-function (pattern var test condition)
  (multiple-value-bind (tests binding-specs)
    (generate-match-body pattern (pattern-free-variables pattern) test)
    `(defun ,(generate-rule-function-name pattern) (p ,@ *bound-vars*)
       (if (and ,@ tests)
	   (values T (list ,@ (if var '(p)) ,@ (reverse binding-specs))
		    ;; return the condition so we know what queue to put the rule on.
		   ',condition)))))

(defun generate-match-body (pattern vars extra-test
			       &aux structure-tests equal-tests
				    binding-specs var-alist)
  (dolist (test (generate-unify-tests pattern vars nil 'P))
    (cond ((variable? (car test))
	   (setq equal-tests (append (generate-pairwise-tests (cdr test))
				     equal-tests))
	   (push (car (last test)) binding-specs)
	   (if test (push (cons (car test) (car (last test))) var-alist)))
	  (t (push test structure-tests))))
  (values (append structure-tests equal-tests
		  (if extra-test
		      (list (sublis var-alist extra-test))))
	  binding-specs))

(defun generate-pairwise-tests (tests)
  (cond ((or (null tests) (null (cdr tests))) nil)
	(t (cons (list 'EQUAL (car tests) (cadr tests))
		 (generate-pairwise-tests (cdr tests))))))

(defun generate-unify-tests (pattern vars tests path)
  (cond ((null pattern) (cons `(null ,path) ;this is the end.
			      (cons `(not (null ,(cadr path)))
				    tests))) ;and it did go out this far.
	((member pattern vars :test #'equal) ;must see if the pattern has been bound
	 ;elsewhere, and if it has, test to see if the element here is
	 ;consistent with that earlier binding.
	 (let ((previous (assoc pattern tests :test #'equal)))
	   (cond (previous ;add this position to test it
		   (setf (cdr previous) (cons path (cdr previous)))
		   tests)
		 (t (cons (list pattern path) tests)))))
	;if variable, it must be bound so test against the current value.
	((variable? pattern) (cons `(equal ,pattern ,path) tests))
	;if not a list, then see if equal (should be number check as well...)
	((not (listp pattern)) (cons `(equal ',pattern ,path) tests))
	(t (generate-unify-tests (cdr pattern) vars
             (generate-unify-tests (car pattern) vars
		(cons `(consp ,path) tests) ;avoid lisp errors
		(list 'car path)) ;extend the path.
             (list 'cdr path))))) ;extend path in other direction

;;;; Recursive macroexpansion
;;   Knowing what forms to skip over is still very implementation
;;   dependent.  It gets hairier when (if) one allow assertions to
;;   have procedural side-effects. (DEBACLE did this to implement
;;   quantifiers properly, and it just didn't seem to be worth it.)

(defun fully-expand-body (body)
  (cond ((null body) nil)
	((not (listp body)) body)
	((symbolp (car body))
	 (case (car body)
	   (QUOTE body)
	   (LAMBDA `(lambda ,(cadr body)
		      ,@ (fully-expand-body (cddr body))))
	   (DO `(do ,(cadr body) ,(caddr body)
		  ,@ (cdddr body)))
	   ((DOTIMES DOLIST LET PROG PROGV FLET LABELS)
	    `(,(car body) ,(cadr body)
	      ,@ (fully-expand-body (cddr body))))
	   ((PROGN PROG1 PROG2)
	    `(,(car body) ,@ (mapcar #'fully-expand-body (cdr body))))
	   (t (let ((new-body (macroexpand body)))
		(cons (fully-expand-body (car new-body))
		      (fully-expand-body (cdr new-body)))))))
	(t (let ((new-body (macroexpand body)))
	     (cons (fully-expand-body (car new-body))
		 (fully-expand-body (cdr new-body)))))))

;;;; Syntactic sugar for rule bodies
;;
;; An "R" prefix on a function indicates that it should
;; automagically expand its arguments to make pattern
;; variables be evaluated at run time.  The most sophisticated
;; version would use *bound-vars* to expand just those with
;; a value, but for simplicity we start by assuming all vars should
;; be evaulated.

(defmacro rnogood (arg &optional (informant ':RNOGOOD))
  `(nogood ,(quotize (mapcar #'desugar-pattern arg)) ,informant))

(defmacro rassert! (arg) `(assert! ,(quotize (desugar-pattern arg))))

(defmacro rassume (arg) `(assume ,(quotize (desugar-pattern arg))))

(defmacro refer-to (arg) `(referent ,(quotize (desugar-pattern arg)) t))

;; Second quote is added to fix a bug in optional arguments to macros
;; (rjustify (foo x) ((bar y)) 'Some-rule)
;;    behaves differently than (rjustify (foo x) ((bar y)))
;;       when optional argument has a single quote

(defmacro rjustify (consequent antecedents &optional (source ''Some-Rule))    
  `(justify ,(quotize (desugar-pattern consequent))
	    ,(quotize (mapcar #'desugar-pattern antecedents)) ,source))

(defvar *quantifiers* '(EVERY SOME))

(defun quotize (pattern &optional (quoted-vars nil))
  (cond ((null pattern) nil)
	((variable? pattern)
	 (if (member pattern quoted-vars) (list 'QUOTE pattern) pattern))
	((atom pattern) (list 'QUOTE pattern))
	 ; :EVAL indicates that whoever built it wants it evaluated at run-time
	((eq (car pattern) ':EVAL) (cadr pattern))
	((eq (car pattern) ':QUOTE) (list 'QUOTE (cadr pattern)))
	((member (car pattern) *quantifiers*)
	 ;; Assume (<quant name> <var or list of vars> . <whatever>)
	 (if (listp (cadr pattern)) (setq quoted-vars (append (cadr pattern) quoted-vars))
	     (setq quoted-vars (cons (cadr pattern) quoted-vars)))
	 `(cons ',(car pattern)
		(cons ',(cadr pattern)
		      ,(quotize (cddr pattern) quoted-vars))))
	(t `(cons ,(quotize (car pattern) quoted-vars)
		  ,(quotize (cdr pattern) quoted-vars)))))

(defmacro rlet (bindings &rest body)
  (let ((*bound-vars* (append (mapcar #'car bindings) *bound-vars*)))
    `(let ,(mapcar #'(lambda (let-clause)
				(list (car let-clause)
				      (if (and (listp (cadr let-clause))
					       (eq (car (cadr let-clause))
						   ':EVAL))
					  (cadr (cadr let-clause))
					  (quotize (cadr let-clause)))))
			    bindings)
       ,@ (fully-expand-body body))))
;; ******* For upward compatibility -- may not be around in next version!
(defmacro nlet (&rest stuff) `(rlet ,@ stuff))

;;;; Running rules

;;; Some statistical counters
(proclaim '(special *winning-matches* *losing-matches*))
(defvar *losing-matches* 0)
(defvar *winning-matches* 0)

(defun insert-rule (class matcher body ins imps)
  ;; Installs a new rule under the appropriate class.
  (let ((rule (make-runk :matcher matcher :body body
			 :in-nodes ins
			 :imp-nodes imps)))
    (push rule (class-rules class)) ;indexing
    (dolist (candidate (class-instances class))
      (try-rule-on rule candidate))))

(defun try-rules (fact)
  ;; Called whenever a new fact is added.
  (dolist (rule (class-rules (assertion-class fact)))
    (try-rule-on rule fact)))

(defun try-rule-on (rule assert)
  ;; Tests the <rule, assert> pairing 
  (multiple-value-bind (match? bindings condition)
      (funcall (runk-matcher rule) (lisp-form assert))
    (cond (match? 
	   (incf *winning-matches*)
	   (case condition
	     (:INTERN
	       (enqueue-intern-rule (runk-body rule) bindings (assertion-class assert)
					   (runk-in-nodes rule)
					   (runk-imp-nodes rule)))
	     (:IN (enqueue-in-rule (runk-body rule) bindings assert (runk-in-nodes rule)
				   (runk-imp-nodes rule)))
	     (:IMPLIED-BY (enqueue-implied-by-rule (runk-body rule) bindings assert
						   (runk-in-nodes rule)
						   (runk-imp-nodes rule)))))
	  (t (incf *losing-matches*)))))

;;;; Debugging routines for matching

(defun test-rule-fact-match (rule assert &optional (prefix "") (stream *standard-output*))
  (multiple-value-bind (match? bindings condition)
      (funcall (runk-matcher rule) (lisp-form assert))
    (cond (match?
	   (format stream "~% ~A Match succeeded; ~A, ~A." prefix bindings condition)
	   t)
	  (t (format stream "~% ~A Match failed." prefix) nil))))

(defun show-matching-facts (rule &optional (stream *standard-output*)
				 &aux count total) 
  (setq count 0 total 0)
  (dolist (fact (class-instances (runk-class rule)))
    (when (test-rule-fact-match rule fact (format nil "~D" count) stream)
      (incf total))
    (incf count))
  (values total count))

(defun show-matching-rules (form &optional (stream *standard-output*)
				 &aux count total fact) 
  (setq count 0 total 0 fact (if (assertion? form) form
				 (referent form t)))
  (dolist (rule (class-rules (assertion-class fact)))
    (when (test-rule-fact-match rule fact (format nil "~D" count) stream)
      (incf total))
    (incf count))
  (values total count))

(defun show-matching-class-entries (class-name &optional (stream *standard-output*)
					       &aux class)
  (setq class (if (class? class-name) class-name (lookup-class class-name t)))
  (dolist (fact (class-instances class))
    (format stream "~% ~A" (lprinc-string (lisp-form fact))) 
    (show-matching-rules fact stream)))

;;;; Queues

;; Go for simplicity here.  Avoid consing by using slot in runk structure
;; itself.  Switch to LIFO for simplicity.

;; An important engineering constraint is to keep the simple cases fast.
;; For instance, QPE uses only :INTERN rules.  Having this other hair
;; should not noticeably slow it down. 

(defun clear-queues ()
  (setq *intern-queue* nil)
  (setq *in-queue* nil)
  (setq *implied-by-queue* nil)
  (setq *implied-by-checked* (cons nil nil))
  (setq *contradiction-consumer-queue* nil))

(defun find-queue-length (runk)
  (labels ((runk-counter (runk length)
	     (cond ((null runk) length)
		   ((null (runk-rest runk)) length)
		   (t (runk-counter (runk-rest runk) (1+ length))))))
    (runk-counter runk 0)))

(defun show-queues (&optional (stream *standard-output*))
  (cond (*intern-queue*
	 (format stream "~% ~D :INTERN rules." (find-queue-length *intern-queue*)))
	(t (format stream "~% :INTERN queue empty.")))
  (cond (*in-queue*
	 (format stream "~% ~D :IN rules." (find-queue-length *in-queue*)))
	(t (format stream "~% :IN queue empty.")))
  (cond (*implied-by-queue*
	 (format stream "~% ~D :IMPLIED-BY rules."
		 (find-queue-length *implied-by-queue*)))
	(t (format stream "~% :IMPLIED-BY queue empty.")))
  (cond (*contradiction-consumer-queue*
	 (format stream "~% ~D contradiction consumers." 
		 (length *contradiction-consumer-queue*)))
	(t (format stream "~% No contradiction consumers.")))
  (format stream "~% ~D winning matches, ~D losing, ~D% hit rate."
	  *winning-matches* *losing-matches*
	  (float (/ (round (* (float (* 100.0 (/ *winning-matches*
					  (+ *winning-matches* *losing-matches*))))
		       10)) 10))))

;; intern rules are still the simplest.

(defun enqueue-intern-rule (body bindings class ins imps)
  (let ((new (make-runk :BODY body :ARGS bindings
			:CLASS class
			:IN-NODES ins :iMP-NODES imps)))
    (setf (runk-rest new) *intern-queue*)
    (setq *intern-queue* new)))

(defun dequeue-intern-rule () 
  (let ((item *intern-queue*))
    (when item
      (setq *intern-queue* (runk-rest item)))
    item))

(defun intern-queue-waiting? () (not (null *intern-queue*)))

;;;;  IN queue manipulations
;;
;; This code is as stupid as possible.  Keeps everything in a linear
;; list, computes labels for collections of trigger nodes from scratch.

(defun enqueue-in-rule (body bindings assert ins imps)
  ;;assert is the newly triggered item
  (let ((item (make-runk :BODY body :ARGS bindings
			 :CLASS (assertion-class assert)
			 :IN-NODES (ordered-cons (get-tms-node assert)
						 ins #'node-order)
			 :IMP-NODES imps)))
    (setf (runk-rest item) *in-queue*)
    (setq *in-queue* item)))

(defun dequeue-in-rule ()
  ;; Goes down queue until it finds an item which is IN.
  ;; Really stupid version, should use discrimination net instead.
  (do ((current *in-queue*)
       (last nil))
      ((null current) nil)
    (cond ((rule-instance-in? current)
	   (cond (last (setf (runk-rest last) (runk-rest current)))
		 (t (setq *in-queue* (runk-rest current))))
	   (return-from DEQUEUE-IN-RULE current))
	  (t (setq last current
		   current (runk-rest current))))))

(defun rule-instance-in? (ri)
  (let ((nodes (runk-in-nodes ri)))
    (cond ((cdr nodes)
	   (do ((envs (tms::label (car nodes)))
		(ns (cdr nodes) (cdr ns)))
	       ((null ns) t)
	     (unless (setq envs
			   (tms::generic-union-envss envs (tms::label (car ns))))
	       (return-from rule-instance-in? nil))))
	  (t ;; single node case
	   (tms::label (car nodes))))))

(defun in-queue-waiting? () (not (null *in-queue*)))

;;;; Implied-by queue

;; A bit more complicated, since we want to avoid redundant work.
;; Once we find a runk isn't implied by the focus, we want to not look
;; at it again until we change the focus.  This means caching the failed
;; runks on the *implied-by-checking* queue, and adding them onto the
;; regular *implied-by-queue* once the focus changes. 

(defun enqueue-implied-by-rule (body bindings assert ins imps)
  (let ((item (make-runk :BODY body :ARGS bindings
			 :CLASS (assertion-class assert)
			 :IN-NODES ins
			 :IMP-NODES (ordered-cons (get-tms-node assert) imps
						  #'node-order))))
    (setf (runk-rest item) *implied-by-queue*)
    (setq *implied-by-queue* item)))

(defvar *imp-max* 0)
(init-form (setq *imp-max* 0))

(defun dequeue-implied-by-rule ()
  ;; Assume focus is set up properly and so on.
  (do ((runk *implied-by-queue*)
       (result nil)
       (counter 0 (1+ counter)))
      ((null runk)
       ;; Debugging
       (if (> counter *imp-max*) (setq *imp-max* counter))
       result)
    (cond ((implied-by-triggers-ready? runk) ;; Got it
	   (setq *implied-by-queue* (runk-rest runk))
	   (setq result runk runk nil))
	  ((car *implied-by-checked*) ;; Add to what is already there
	   (setq *implied-by-queue* (runk-rest runk))
	   (setf (runk-rest runk) (car *implied-by-checked*))
	   (setf (car *implied-by-checked*) runk)
	   (setq runk *implied-by-queue*))
	  (t ;; Initialize offloading
	   (setq *implied-by-queue* (runk-rest runk))
	   (setf (runk-rest runk) nil)
	   (setf (car *implied-by-checked*) runk)
	   (setf (cdr *implied-by-checked*) runk)
	   (setq runk *implied-by-queue*)))))

(defun implied-by-triggers-ready? (runk)
  (every #'(lambda (node) (tms::true-in? node *focus*)) (runk-imp-nodes runk)))

(defun implied-by-queue-waiting? () (and (focus-okay?) (not (null *implied-by-queue*))))

(defun implied-by-queue-length ()
  (do ((runk *implied-by-queue*
	     (runk-rest runk))
       (counter 0 (1+ counter)))
      ((null runk) counter)))

(defun revert-implied-by-queue ()
  ;; Two things need to happen: First, the rules cached under *implied-by-checked*
  ;; need to be moved back.  Second, any rules on the in-queue which have implied-by
  ;; triggers need to be put back on the implied-by queue to be checked again.
  ;; Otherwise, implied-by triggers won't necessarily be correctly enforced.
 (when (car *implied-by-checked*) ;; Got some cached
   (setf (runk-rest (cdr *implied-by-checked*)) *implied-by-queue*)
   (setq *implied-by-queue* (car *implied-by-checked*))
   (setf (car *implied-by-checked*) nil)
   (setf (cdr *implied-by-checked*) nil))
 (do ((current *in-queue*)
      (last nil))
     ((null current))
   (cond ((runk-imp-nodes current)
	  (cond (last  ;; Splice it out
		 (setf (runk-rest last) (runk-rest current))
		 (setf (runk-rest current) *implied-by-queue*)
		 (setq *implied-by-queue* current)
		 (setq current (runk-rest last)))
		(t ;; Move it, and reset the top
		 (setq *in-queue* (runk-rest current))
		 (setf (runk-rest current) *implied-by-queue*)
		 (setq *implied-by-queue* current)
		 (setq current *in-queue*))))
	 (t (setq last current
		  current (runk-rest current))))))

;;;; Contradiction handlers

(defvar *contradiction-consumer-queue* nil)

(defun contra-queue-waiting? () *contradiction-consumer-queue*)

(defun add-contradiction-consumer (env procedure)
  (let ((entry (assoc env *contradiction-consumer-queue*)))
    (unless entry
      (push (setq entry (cons env nil)) *contradiction-consumer-queue*))
    (push (list procedure (list env)) (cdr entry))))

(defun dequeue-contra-rule (&aux rule)
  (dolist (entry *contradiction-consumer-queue*)
;; Make it implied by focus??
    (when (tms::env-contradictory (car entry))
      (setq rule (pop (cdr entry)))
      (if (null (cdr entry))
	  (setq *contradiction-consumer-queue*
		(delete entry *contradiction-consumer-queue*)))
      (if rule (return-from dequeue-contra-rule rule)))))

;;;; Running rules
(proclaim '(special **RULE-SELF**))

(defun run-rule (rule)
  (when rule
    (setq **RULE-SELF** rule)
      ;(if-debug-flag show-rule (format T "~%Running rule:") (g rule))
    (let ((*IN-NODES* (runk-in-nodes rule))
	  (*IMP-NODES* (runk-imp-nodes rule)))
      (apply (runk-body rule) (runk-args rule)))))

(defvar *Number-of-Rules-Run* 0) ;; Statistics
(defvar *contra-rules-run* 0) ;; Yet more statistics

(defun run ()
  (do ((number-of-rules 0 (if quiescent? number-of-rules
			      (1+ number-of-rules)))
       (rule nil) (quiescent? nil))
      (quiescent?
       (if-debug-flag rules			
	  (format t "~%    ADB: ~D rules run." number-of-rules)) 
       (setq *number-of-rules-run*
	     (+ number-of-rules *number-of-rules-run*))
       number-of-rules)
    (setq quiescent? t)
    (when (intern-queue-waiting?)
      (setq quiescent? nil
	    rule (dequeue-intern-rule)) 
     (if-debug-flag rules
		    (format t "~% Got intern rule, ~A" rule))
      (run-rule rule))
    (when (and quiescent? (contra-queue-waiting?))
      (setq rule (dequeue-contra-rule))
      (when rule
	(if-debug-flag rules
		       (format t "~% Got contradiction handler, ~A." rule))
	(setq quiescent? nil) (incf *contra-rules-run*)
	(run-rule rule))) 
    (when (and quiescent? (implied-by-queue-waiting?))
      (setq rule (dequeue-implied-by-rule))
      (when rule
	(if-debug-flag rules
		       (format t "~% Got implied-by rule, ~A" rule))
	(setq quiescent? nil)
	(cond ((runk-in-nodes rule) ;; Got to check the IN conditions next
	       (setf (runk-rest rule) *in-queue*)
	       (setq *in-queue* rule))
	      (t ;; otherwise, just do it
	       (run-rule rule)))))
    (when (and quiescent? (in-queue-waiting?))
      (setq rule (dequeue-in-rule)) ;; can still lose due to rule not being in
      (when rule (setq quiescent? nil)
	   (if-debug-flag rules
	     (format t "~% Got intern rule, ~A" rule))
	   (cond ((runk-imp-nodes rule) ;; Got to check focus
		  (cond ((and (focus-okay?) (implied-by-triggers-ready? rule))
			 (run-rule rule))
			(t (setf (runk-rest rule) *implied-by-queue*)
			   (setq *implied-by-queue* rule))))
		 (t (run-rule rule)))))))

;;;; Focus mechanism
;;
;; The variable *focus* is interpreted as what the problem solver is thinking about
;; right now.
;;
;; The :IMPLIED-BY condition for rules means the body will only be executed
;;  when the triggers are implied by the current focus.  Checking for this condition
;; is quite cheap, since it can be checked for each trigger independently.
;;
;; Another possible condition is :CONSISTNT-WITH, meaning the body will only be
;; executed when the triggers can all be consistently believed with the focus environment.
;; Such rules would be a specialization of :IN rules, and thus quite expensive (i.e., intermediate
;; results of indeterminate size could be generated).  It isn't too clear what these rules could really
;; do that :IMPLIED-BY wouldn't be better for, so this condition is not implemented now.

(defun focus-okay? ()
  (and *focus* (not (tms::env-contradictory *focus*))))

(defun change-focus (new-focus)
  ;; notify TMS when hook is there
  (unless (tms::env-contradictory new-focus)
    (setq *focus* new-focus)
    (revert-implied-by-queue)))

(defmacro With-focus (focus &rest body)
  `(let ((old-focus *focus*))
     (unless (tms::env-contradictory ,focus)
       (change-focus ,focus)
       ,@ body
       (change-focus old-focus))))

;; ****** Find the appropriate tms:: replacement for this
(defun node-order (n1 n2) (< (tms::node-unique n1) (tms::node-unique n2)))
