;;; -*- Mode: LISP; Syntax: Common-lisp; Package: ADB -*-

(in-package 'ADB)

;;;; Copyright (c) 1989, Kenneth D. Forbus, University of Illinois
;;;; All rights reserved.

;;;; Closed-world set mechanism

;;; Syntax:
;;; (<set> members <elements>) means that <set> = <elements>.
;;; (<set> has-member <element>) means that <element> is in <set>.

;;; The idea of closed-world sets is that sometime in processing,
;;; the labels for all the HAS-MEMBER statements are maximal.  This
;;; means that one can compute, once and for all, what possible constituents
;;; the set has.  These are called the "construals" of the set.

;;; (CLOSE-SET <set>) generates all the construals by closing the set.
;;; It does NOT make an explicit closed-world assumption, since we are
;;; presuming maximal labels for the HAS-MEMBER statements.  (Note to builders
;;; of incremental systems: You'll need explicit CWA's!)

;;; The algorithm of CLOSE-SET is simple:
;;; 1. Gather membership statements.
;;; 2. Close the positive membership statements.  That is, ensure that whenever they
;;;    are not known to be true, their negation is believed.
;;; 3. Construct construals by a form of interpretation construction.

(defun close-set (set &aux elements poss negs choices choice)
  (dolist (form (fetch `((,set HAS-MEMBER ?EL) . ?X)))
    (pushnew (first form) elements :test #'equal))
  (dolist (el elements) (close-fact el :TRUE))
  ;; Special cases: some elements might always be there, others
  ;; might never be.  When computing the choice sets, initialize with
  ;; these.
  (dolist (el elements)
    (setq choice (cons (cons el :TRUE) (cons el :FALSE)))
    (cond ((true? (car choice)) (push (car choice) poss))
	  ((true? (cdr choice)) (push (cdr choice) negs))
	  (t (push choice choices))))
  (build-set-construals set (list tms::*empty-env*) choices poss negs))

(defun build-set-construals (set envs choices poss negs &aux new-envs)
  (cond ((null choices)
	 (cond ((or poss negs) ;; Install justification
		(justify `((,set MEMBERS ,(mapcar #'(lambda (x) (third (car x)))
						  poss)) . :TRUE)
			 (append poss negs) :SET-CWA))
	       (t ;; Very special case -- nothing known.
		;; Should there be an error detection rule here??
		(assert! `((,set MEMBERS NIL) . :TRUE)))))
	(t ;; Recurse down both alternatives
	 (setq new-envs (tms::generic-union-envss (environments (caar choices)) envs))
	 (when new-envs
	   (build-set-construals set new-envs (cdr choices) (cons (caar choices) poss) negs))
	 (setq new-envs (tms::generic-union-envss (environments (cdar choices)) envs))
	 (when new-envs
	   (build-set-construals set new-envs (cdr choices) poss (cons (cdar choices)
								       negs))))))

(defForm-Parser-cadr (?set MEMBERS ?elements)
		     (adb::parse-form (list '= ?set (cons 'set-members ?elements))))

(defForm-Parser (set-members . ?members)
		(nconc (list :START (if (eq adb::*report-mode* :LATEX) "\\{" "{"))
			(adb::parse-argument-list ?members)
			(list (if (eq adb::*report-mode* :LATEX) "\\}" "}") :END)))

(defform-parser-cadr (?set has-member ?el)
		     (adb::parse-form (list 'has-member ?set ?el)))

;;;; Testing the set mechanism.

(defun test-cwa-set1 ()
  (adb::init)
  (assumep 'a1)
  (assumep 'a2)
  (assumep 'a3)
  (assert! '(A4 . :TRUE))
  (justify '((foo has-member 1) . :TRUE) '((a1 . :TRUE) (a2 . :TRUE)))
  (justify '((foo has-member 2) . :TRUE) '((a1 . :TRUE) (a2 . :FALSE)))
  (justify '((foo has-member 3) . :TRUE) '((a3 . :TRUE) (a4 . :TRUE)))
  (justify '((foo has-member 4) . :TRUE) '((a3 . :TRUE) (a4 . :FALSE)))
  (close-set 'foo))