;;; -*- Mode: LISP; Syntax: Common-Lisp; Package: ADB; -*-

(in-package 'adb)

;;;; ATMoSphere toplevel
;;

(export '(init-form after-init-form *current-adb* init get-solutions
	  print-statistics print-listed-statistics list-statistics
	  contradiction-distribution print-contradictions print-envs
	  print-env okay-packet? in-packet packet-end use-packet
	  packet-defined? eval-or-packetize store-in-packet
          load-in-packet load-forms-into-packet add-to-end-of-packet
          clear-packet merge-packets add-to-packet clear-all-packets
          show-packets show-packet packet-contents))

(proclaim '(special *fact-hash-table* *fact-table-size*
		  *class-hash-table* *class-table-size*
		  *queue *number-of-rules-run*
		  *packets* *default-packet*
		  *number-of-classes* *assert*
		  *end-forms*))

(defmacro Init-Form (form)
 ;Init forms will always be called at least once.
 `(progn (setq *init-forms*
	       (cons ',form (delete ',form *init-forms*
					      :test #'EQUAL)))))

(defmacro After-Init-Form (form)
  ;Init forms will always be called at least once.
  `(progn (setq *after-init-forms*
		(cons ',form (delete ',form *after-init-forms*
					      :test #'EQUAL)))))

(defmacro Init-Once-Form (form)
  ;; Executed once and then flushed.  This is for getting rid of
  ;; example-specific information.
  `(progn (push ',form *init-once-forms*)))

(defvar *init-forms* nil "Forms to run when initializing")
(defvar *after-init-forms* nil "Forms to run after initializing")
(defvar *init-once-forms* nil "Executed once then flushed")

;;;; Initialization

(defvar *Current-Adb* (cons nil nil)
  "Flag to let subsystems know the world has changed out from under them.")

(proclaim '(special *focus*))

(defun init (&optional (reset? nil))
  ;;This must be called before doing anything
  (setq *current-Adb* (cons nil nil))
  (tms:init-tms) ;; Johan says to leave the optional arg up to him.
  (clear-tables reset?)
  (clear-counters)
  (clear-queues)
  (dolist (form *init-once-forms*) (eval form))
  (setq *init-once-forms* nil)
  (dolist (form *init-forms*) (eval form))
  (setq *focus* tms::*empty-env*)
  (dolist (form *after-init-forms*) (eval form))
  (use-packet :ATMoSphere-basics))

(proclaim '(special *winning-matches* *losing-matches*
		    *contra-rules-run* *rule-name-counter*))

(defun clear-counters ()
  (setq *number-of-rules-run* 0)
  (setq *number-of-classes* 0)
  (setq *assert* 0)
  (setq *winning-matches* 0
	*losing-matches* 0)
  (setq *end-forms* nil)
  (setq *contra-rules-run* 0)
  (setq *rule-name-counter* 0))

;;; User interface functions

(proclaim '(special *solutions*))

(defun get-solutions (&optional (print? t) (classes tms:*classes*))
  (setq *solutions* (solutions classes))
  (when print?
    (format t "~%~A solutions" (length *solutions*))
    (dolist (s *solutions*)
      (format t "~%~A" (tms:string-env (car s)))
      (read-char))))

(defun contradiction-distribution (&optional (stream *terminal-io*) &aux result total)
  (format stream "~%Length   # nogoods")
  (multiple-value-setq (result total) (tms::nogood-distribution))
  (dolist (entry (nreverse result))
    (format stream "~%~5<~D~>~12<~D~>" (car entry) (cdr entry)))
  (format stream "~% ~D nogoods in all" total))

(defun print-contradictions (from to &optional (break? t))
  (do ((i from (1+ i))
       (counter 0))
       ((> i to) counter)
      (dolist (e (tms::collect-minimal-nogoods i))
	(incf counter)
	(print-env e)
	(if break? (read-char)))))

(defun contradictions-involving
       (asn &optional (from 2) 
	    (to (apply #'max
		       (mapcar #'car
			       (tms::nogood-distribution))))
	    &aux nogoods)
  (do ((i from (1+ i)))
      ((> i to) nogoods)
    (dolist (e (tms::collect-minimal-nogoods i))
      (when (member asn (tms::env-assumptions e))
	(push e nogoods)))))

(defun nogoods-in-tree-entry (entry)
  (let ((*nogoods-this-entry* nil))
    (labels ((nogoods-in-subtree (entry)
	       (when entry
		 (cond ((listp entry)
			(cond ((listp (cdr entry))
			       (dolist (subentry (cdr entry))
				 (nogoods-in-subtree subentry)))
			      (t (push (cdr entry) *nogoods-this-entry*))))
		       (t (push (cdr entry) *nogoods-this-entry*))))))
      (nogoods-in-subtree entry)
      *nogoods-this-entry*)))

;;;;; Finding causes for nogoods
;;
;; A common problem when debugging ATMS code is that some list of facts, which 
;; should be consistent, isn't.  These functions help interrogate the nogood database.

(defun map-nogoods (fun) (tms::walk-contradictions fun))

(defun nogoods-including (facts)
  (nogoods-including-assumptions (mapcar #'fact-assumption facts)))

(defun nogoods-including-assumptions (asns)
  (let ((*nogoods-found* nil))
    (map-nogoods #'(lambda (nogood)
		     (if (every #'(lambda (asn) 
				    (member asn (tms::env-assumptions nogood)))
				asns)
			 (push nogood *nogoods-found*))))
    *nogoods-found*))

;;;; Packet subsystem

;; The idea is to insulate the system from having to re-load files all the time.
;; This is accomplished by creating a cache of PACKETS, which contain forms that
;; index rules, add data, or whatever.
;; Each rule packet consists of a name and a list of indexing forms.  Evaluating
;; the list of indexing forms will re-install the rules, data, or whatever.
;; The name of a packet must be a symbol.

;; The following operations are supported on packets:
;;(USE-PACKET <name>) causes the rules to be installed in the current incarnation
;; of the system.  If you want to have some rule packet appear in every incarnation,
;; add this form to the after-init forms.
;; (CLEAR-PACKET <name>) wipes out an existing rule packet.
;; (SHOW-PACKETS &optional stream) lists the rule packets and some statistics about them.
;; (IN-PACKET <name>) Sets the default rule packet to be <name>.  This means
;; any rules loaded will be added to that rule packet, until RULE-PACKET-END is called. 
;; (RULE-END) Marks the end of a set of rules to be loaded into the default packet.
;; It is legal to add more rules to a packet later, that is, one can call IN-RULE-PACKET
;; an arbitrary number of times with the same name to build up a packet of rules from multiple
;; files.
;; (MERGE-PACKETS <result> <name1> ... <namen>)  Creates a new rule packet <result>
;; by appending rule packets <name1> ... <namen>.  

;;; Intended for internal use:
;; (OKAY-PACKET? <name>) returns non-nil if <name> could potentially be a packet.

(defvar *packets* nil)
(defvar *default-packet* nil)

(defun okay-packet? (pkt) (and pkt (symbolp pkt)))

(defun in-packet (name)
  (if (okay-packet? name)
      (setq *default-packet* name)
      (error "Rule packet name must be a symbol: ~A." name)))

(defun packet-end () (setq *default-packet* nil))

(defun use-packet (name)
  (unless (okay-packet? name)
    (error "Rule packet name must be a symbol: ~A." name))
  (let ((packet (assoc name *packets*)))
    (when packet
      (dolist (form (cdr packet)) (eval form)))))

(defun packet-defined? (pkt)
  (when (okay-packet? pkt)
    (let ((stuff (assoc pkt *packets*)))
      (cond ((and stuff (cdr stuff)) (length (cdr stuff)))
	    (t nil)))))

(defun eval-or-packetize (form)
  `(cond ((okay-packet? *default-packet*)
	  (add-to-packet *default-packet* ',form))
	 (t ,form)))

;;;; Packet system, continued.

(defmacro Store-In-Packet (pkt &rest forms)
  (unless (okay-packet? pkt)
    (error "Bad packet name: ~A."))
  `(progn (let ((*default-packet* ',pkt))
	    ,@ forms)))

(defun load-in-packet (file pkt &optional (package adb::*user-package*))
  (let ((*package* *package*))
    (let ((*default-packet* pkt)
	  (*package* (if (packagep package) package
			 (find-package package))))
      (load file))))

(defun load-forms-into-packet (file pkt &optional (package adb::*user-package*))
  ;; This is for stuff that should be eval'ed at run-time.
  (with-open-file (stream file :direction :input)
    (let ((marker (cons nil nil))
	  (*package* (if (packagep package) package
			 (find-package package))))
      (do ((form (read stream  nil marker)
		 (read stream  nil marker))
	   (forms nil))
	  ((eq form marker)
	   (add-to-end-of-packet (nreverse forms) pkt))
	(push form forms)))))

(defun add-to-end-of-packet (new-stuff pkt)
  (unless (okay-packet? pkt)
    (error "ATMoSphere packet name must be a symbol: ~A." pkt))
  (let ((stuff (assoc pkt *packets*)))
    (unless stuff
      (push (setq stuff (cons pkt nil)) *packets*))
    (setf (cdr stuff)
	  (nconc (cdr stuff) new-stuff))))

(defun clear-packet (name)
  (unless (okay-packet? name)
    (error "ATMoSphere packet name must be a symbol: ~A." name))
  (let ((packet (assoc name *packets*)))
    (when packet
      (setq *packets* (delete packet *packets*))))
  nil)

(defun merge-packets (result &rest packets)
  (clear-packet result)
  (let ((stuff nil))
    (dolist (packet packets)
      (let ((old-stuff (assoc packet *packets*)))
	(unless old-stuff
	  (error "Not a rule packet: ~A." packet))
	(setq stuff (append stuff (cdr old-stuff)))))
    (push (cons result stuff) *packets*)))

(defun add-to-packet (name thing)
  (let ((packet (assoc name *packets*)))
    (unless packet
      (push (setq packet (cons name nil)) *packets*))
    (push thing (cdr packet))))

(defun clear-all-packets () (setq *packets* nil))

;;;; Debugging packets

(defun show-packets (&optional (stream *standard-output*) &aux counter)
  (setq counter 0)
  (dolist (packet *packets*) (incf counter)
    (format stream "~% ~A: ~D items." (car packet) (length (cdr packet))))
  (format stream "~% ~D packets total." counter))

(defun show-packet (pkt &optional (stream *standard-output*))
  (cond ((not (okay-packet? pkt))
	 (format stream
		 "~A not a valid ATMoSphere packet name -- must be non-nil symbol." pkt))
	(t (let ((stuff (assoc pkt *packets*)))
	     (cond (stuff 
		    (format stream "For packet ~A:" pkt)
		    (let ((*print-length* 5)
			  (*print-level* 3))
		      (dolist (item (cdr stuff))
			(pprint item stream))))
		   (t (format stream "Packet ~A is empty." pkt)))))))

(defun packet-contents (pkt)
  (cond ((not (okay-packet? pkt))
	 (format stream
		 "~A not a valid ATMoSphere packet name -- must be non-nil symbol." pkt))
	(t (cdr (assoc pkt *packets*)))))

;;;; Printing statistics

(proclaim '(special *contra-rules-run* *justify*))

(defun print-statistics (&optional (stream *terminal-io*))
  (format stream "~% ADB STATISTICS")
  (format stream "~% # database items: ~D" *assert*)
  (format stream "~% # database classes: ~D" *class*)
  (format stream "~% # rules run: ~D" *number-of-rules-run*)
  (format stream "~% # C-rules run: ~A" *contra-rules-run*)
  (format stream "~% # TMS nodes: ~D" (1- tms:*node-counter*))
  (format stream "~% # TMS classes: ~D" (length tms:*classes*))
  (format stream "~% # Nogoods: ~D" (1- tms:*contra-counter*))
  (format stream "~% # assumptions: ~D" (1- tms:*assumption-counter*))
  (format stream "~% # environments: ~D" (1- tms:*env-counter*))
  (format stream "~% # Justifications: ~D" *justify*)
  (format stream "~% Distribution of Nogoods:")
  (contradiction-distribution stream))

(defun list-statistics ()
  (list *assert* *class*
   *number-of-rules-run* *contra-rules-run* (1- tms:*node-counter*)
   (length tms:*classes*) (1- tms:*contra-counter*)
  (1- tms:*assumption-counter*) (1- tms:*env-counter*)
  *justify*
  (nogood-distribution-list)))

(defun print-listed-statistics (stats &optional (stream *terminal-io*))
  ;; Prints a list of statistics
  (format stream "~%# database items: ~D" (nth 0 stats))
  (format stream "~%# database classes: ~D" (nth 1 stats))
  (format stream "~%# rules run: ~D" (nth 2 stats))
  (format stream "~%# C-rules run: ~D" (nth 3 stats))
  (format stream "~%# TMS nodes: ~D" (nth 4 stats))
  (format stream "~%# TMS classes: ~D" (nth 5 stats))
  (format stream "~%# Nogoods: ~D" (nth 6 stats))
  (format stream "~%# Assumptions: ~D" (nth 7 stats))
  (format stream "~%# Environments: ~D" (nth 8 stats))
  (format stream "~%# Justifications: ~D" (nth 9 stats))
  (format stream "~%Distribution of Nogoods:")
  (print-nogood-distribution-list (nth 10 stats) stream))

(defun nogood-distribution-list ()
  (multiple-value-bind (result count) (tms::nogood-distribution)
    (cons count (nreverse result))))

(defun print-nogood-distribution-list (almost-alist stream)
  (format stream "~%Length   # nogoods")  
  (dolist (entry (cdr almost-alist))
    (format stream "~%~5<~D~>~12<~D~>" (car entry) (cdr entry)))
  (format stream "~% ~D nogoods in all" (car almost-alist)))

;;;; Random print stuff

(defun print-envs (fact &optional (break? t) (stream *terminal-io*))
  (format stream "~%For ")
  (lprinc fact stream)
  (format stream ":")
  (dolist (e (tms::label (get-tms-node fact)))
    (print-env e stream)
    (if break? (read-char)))
  fact)

(defun e (env) (tms:e env))

(defun print-env (env &optional (stream *terminal-io*))
  (format stream "~% Env ~A:" env)
  (if (tms:env-contradictory env) (format stream "*"))
  (dolist (as (tms:env-assumptions env))
    ;probably wrong in general
    (format stream "~%  ")
    (lprinc (view-node (get-assumption-node as)) stream))
  env)

(defun show-envs (envs)
  (format t "~% There are ~D environments.~%" (length envs))
  (dolist (e envs)
    (print-env e)
    (unless (y-or-n-p "~%Continue?")
      (return nil))))

(defun run-test-file (file &optional (break? nil))
  ;;;; This is mainly for debugging.
  (with-open-file (input file ':direction ':IN)
    (let ((marker (cons nil nil)))
      (do ((form (read input marker)
		 (read input marker)))
	  ((eq form marker) t)
	(if-debug-flag input (pprint form))			
	(eval form)
	(let ((num (run)))
	  (if-debug-flag input (print num)))
	(if break? (break file t))))))

