;;; -*- Mode: LISP; Syntax: Common-lisp; Package: ADB -*-

;;;; Interface goodies for ATMoSphere

;; Copyright (c) 1989, Kenneth D. Forbus, University of Illinois
;; All rights reserved.

(in-package 'adb)

;;;; Interogatives

(defun say-indent (indent)
  (format t "~%")
  (dotimes (i indent) (format t " ")))

(defun say-justification (just &optional (indent 3.))
  (format t "~%~% Informant = ~A" (car just))
  (dolist (node (cdr just))
    (say-indent indent)
    (cond ((tms::node? node)
	   (format t "~A: ~A"
		   (lprinc-string (view-node node)) (tms::label node)))
	  (t (format t "** assumption ~A **" (tms::assumption-unique node))))))

(defmacro whyq (thing) `(why ',thing))
(defmacro why-notq (thing) `(why-not ',thing))

(defun why (fact &aux node justifications)
  (setq node (get-tms-node (desugar-pattern fact)))
  (format t "~%~A~%  is ~A, Label = ~A"
	  (lprinc-string (view-node node))
	  (tms::status-of-node node)
	  (tms::label node))
  (if (tms::node-assumption node)
    (format t "~% It is an assumption.")
    (format t "~% It is not an assumption."))
  (when (setq justifications (tms::node-justifications node))
    (format t "~% Justifications are:")
    (dolist (just justifications) (say-justification just)))
  (unless justifications (format t "~% No known justifications"))
  node)

(defun justifications (fact &optional (stream *terminal-io*) &aux node)
  (setq node (get-tms-node (desugar-pattern fact)))
  (format stream "~%Justified by:")
  (dolist (just (tms::node-justifications node))
    (print-justification just stream))
  (unless (tms::node-justifications node) (terpri stream)))

(defun print-justification (just &optional (stream *terminal-io*))
  (format stream "~%Using ~A on" (car just))
  (dolist (node (cdr just))
    (format stream "~%     ~A" (lprinc-string (view-node node)))))

(defun why-not (fact &aux node nogoods)
  (setq node (get-tms-node (desugar-pattern fact)))
  (format t "~%~A~%  is ~A" (lprinc-string (view-node node))
	  (tms::status-of-node node))
  (if (tms::node-assumption node)
      (format t "~% It is an assumption.")
      (format t "~% It is not an assumption."))
  (when (setq nogoods (remove-if-not #'(lambda (x) (eq (car x) tms::*contra-node*))
				  (tms::node-consequents node)))
    (format t "~% Nogoods are:")
    (dolist (ng nogoods) (say-justification (cdr ng))))
  (unless nogoods (format t "~% No known nogoods."))
  node)

(defun consequences (fact &optional (only-when-holds? t) (break? nil)
			  &aux node conseqs)
  (setq node (get-tms-node (desugar-pattern fact)))
  (format t "~%~A~%  is ~A" (lprinc-string (view-node node))
	  (tms::status-of-node node))
  (if (tms::node-assumption node)
      (format t "~% It is an assumption.")
      (format t "~% It is not an assumption."))
  (when (setq conseqs (tms::node-consequents node))
    (format t "~% Direct consequences are:")
    (dolist (ng conseqs) 
      (when (or (not only-when-holds?)
		(tms::label (car ng)))
      (format T "~%~%~A, by:" (lprinc-string (view-node (car ng))))
      (say-justification (cdr ng))
      (if break? (read-char)))))
  (unless conseqs (format t "~% No known consequences."))
  node)

(defun possible-values? (fact &optional (env nil) &aux node)
  (format t "~%For ~A:" (adb::lprinc-string fact))
  (dolist (val (fetch (cons fact '?x) :condition nil :bindings '(?x)))
    (setq node (get-tms-node (cons fact (car val))))
    (format t "~%  ~A is ~A" (adb::lprinc-string (view-node node))
	    (if env (tms::possible-status-of-node node env)
		    (tms::label node)))))

(defun pvs (f &optional (env nil)) (possible-values? f env))

;;;; More interogatives

(defun compatible-facts? (f1 f2)
  (tms::consistent-nodes? (list (get-tms-node f1) (get-tms-node f2))))

(defun fact-consistent-in? (fact env)
  (tms::consistent-in? (get-tms-node fact) env))

(defun fact-consistent-in-any? (fact env-list)
  (tms::consistent-in-any? (get-tms-node fact) env-list))

(defun tms::consistent-in-any? (node env-list)
  (dolist (env env-list)
    (if (tms::consistent-in? node env) (return t)))) 

(defun implied-by? (fact env) (tms::true-in? (get-tms-node fact) env))

(defun tms::implied-by-any? (node env-list)
  (dolist (env env-list) (if (tms::true-in? node env) (return t))))

(defun in? (fact) (tms::in? (get-tms-node fact)))

(defun out? (fact) (tms::out? (get-tms-node fact)))

(defun true? (fact) (tms::true? (get-tms-node fact)))

(defun false? (fact) (tms::false? (get-tms-node fact)))

(defun fact-status (fact) (tms::status-of-node (get-tms-node fact)))

(defun environments (fact) (tms::label (get-tms-node fact)))

(defun fact-nogoods (fact)
  (cond ((fact-assumption? fact) (tms::assumption-nogoods (fact-assumption fact)))
	(t nil)))

(defun facts-nogoods (facts &aux result)
  ;; Assumes all facts are assumptions
  (setq result (fact-nogoods (car facts)))
  (dolist (fact (cdr facts))
    (setq result (intersection result (fact-nogoods fact)))
    (unless result (return-from facts-nogoods NIL)))
  result)

;;;; Yet more interrogatives

(defun environment-facts (env)
  ;; returns list of facts corresponding to the env.
  (mapcar #'(lambda (asn) (view-assumption asn)) (tms::env-assumptions env)))

(defun assume-fact-if-needed (fact)
  (unless (fact-assumption? fact) (assume! fact))
    (fact-assumption fact))

(defun fact-assumption? (fact) (tms::node-assumption (get-tms-node fact)))

(defun fact-assumption (fact) (tms::node-assumption (get-tms-node fact)))

(defun fact-asn-env (fact) (tms::assumption-env (fact-assumption fact)))

(defun create-environment (facts) ;must all be assumptions
  (unless (every #'fact-assumption? facts)
    (error "~% Must be assumptions - ~A" facts))
  (tms::find-or-make-env (mapcar #'(lambda (x) (fact-assumption? x)) facts)))

(defun environment-cons (fact env) (tms::cons-env env (fact-assumption? fact)))

(defun combine-facts-with-envs (facts envs)
  (tms::generic-union-envss (reduce #'tms::generic-union-envss
				    (mapcar #'environments (cdr facts))
				    :initial-value (environments (car facts)))
		    envs))

(defun lookup-fact (fact env &aux r)
  (setq r (referent fact))
  (if r
    (if env (if (implied-by? r env) (values fact nil))
	(if (in? r) (values fact (environments fact))
	    nil))
    nil))

(defun assumption-order (a1 a2) 
  (< (tms::assumption-unique a1) (tms::assumption-unique a2)))

(defun truthify (fact)
  (if (and (consp fact)
	   (or (eq (cdr fact) ':TRUE)
	       (eq (cdr fact) ':FALSE)))
      fact
      (cons fact ':TRUE)))

;;;; Checking consistency

(defun test-pairwise-compat (the-nodes &optional (break? nil))
  (do ((node (car the-nodes) (car nodes))
       (nodes (cdr the-nodes) (cdr nodes)))
      ((null nodes) nil)
    (do ((others nodes (cdr others)))
	((null others))
      (test-node-pair-compat node (car others))
      (if break? (read-char)))))

(defun test-node-pair-compat (node1 node2)
  (format t "~%1. ~A~%2. ~A~% ~A"
	  (view-node node1)
	  (view-node node2)
	  (tms::consistent-nodes? (list node1 node2))))

(defun diff-env (&rest ignore) nil) ;really useful

;; Attempt to find which contradiction killed the union
 
(defun find-fact-union-killer (f1 f2)
  (declare (ignore ignore))
  (multiple-value-bind (ignore result)
      (tms::consistent-nodes? (list (get-tms-node f1) (get-tms-node f2)))
    result))

;;;; Examining sets of envs

(defun analyze-envs-set (envs)
  (setq *base-env* (envs-common-env envs))
  (setq *residue-envs* (mapcar #'(lambda (in)
				   (env-residue-env in *base-env*))
			       envs))
  (list *base-env* *residue-envs*))

(defun envs-common-env (envs)
  (let ((result (envs-common-assumptions envs)))
    (if result (tms::find-or-make-env result)
	nil)))

(defun envs-common-assumptions (envs &aux common)
  ;returns an environment which represents the
  ;common assumptions of the set of environments
  (when envs
    (setq common (tms::env-assumptions (car envs)))
    (dolist (e (cdr envs))
      (setq common (intersection (tms::env-assumptions e) common))))
  common)

(defun env-residue-env (env base-env)
  (let ((result (env-residue env (tms::env-assumptions base-env))))
    (if result (tms::find-or-make-env result) nil)))

(defun env-residue (env base-asns)
  (remove-if #'(lambda (as) (member as base-asns)) (tms::env-assumptions env)))

(defun find-subsuming-contradiction (env)
  ;finds the simplest contradictions which wipe out the given
  ;contradictory env.  Find all of same simplicity order instead of
  ;just one in order to catch multiple bugs.  I doubt that there is much
  ;use in going to higher levels, since most of them will just be subsumption
  ;of lower ones.  But maybe that should be looked at, too.
  (cond ((null (tms::env-contradictory env))
	 (format t "~%  ~A isn't contradictory." env))
	(t (do ((assumptions (tms::env-assumptions env))
		(counter (length (tms::env-assumptions env)) (1- counter))
		(last-layer (cons tms::*empty-env* nil) envs)
		(envs nil nil)
		(losers nil))
	       ((= counter 0) losers)
	     (do ((assumes (cdr assumptions) (cdr assumes))
		  (assume (car assumptions) (car assumes))
		  (env nil))
		 ((null assume))
	       (dolist (old-env last-layer)
		 (unless (member assume (tms::env-assumptions old-env)) ;avoid redundant overlap
		   (setq env (tms::cons-env old-env assume))
		   (cond ((member env envs)) ;already found it
			 ((member env losers)) ;already found loser
			 ((tms::env-contradictory env) ;a new loser
			  (push env losers))
			 (t (push env envs)))))))))) ;record it and go on.

;;;; Fetching alternate assumptions
;;
;;A common problem in algorithms that manipulate sets of assumption is
;;modelling change by subtracting some assumptions and adding others.
;;Here are some utilities for making this easier.

(defun assumption-alternates (as)
  (mapcar #'(lambda (n) (tms::node-assumption n))
	  (tms::class-nodes (car (tms::node-classes
				 (get-assumption-node as))))))

(defun tms::put-env-prop (env val prop)
 (setf (getf (tms::env-datum env) prop)  val))

(defun tms::get-env-prop (env prop)
  (getf (tms::env-datum env) prop))

(defun tms::put-assumption-prop (as val prop)
  (unless (and (listp (tms::assumption-datum as))
	       (eq (car (tms::assumption-datum as)) 'plist))
    (when (tms::assumption-datum as)
      (format t "~%Someone is using the datum slot after all... ~A." as)
      (error "~%Grumble!"))
    (setf (tms::assumption-datum as) (cons 'plist nil)))
  (setf (get (tms::assumption-datum as) prop)  val))

(defun tms::get-assumption-prop (as prop)
  (get (tms::assumption-datum as) prop))

;;;; Finding NOGOODS a node participates in
;; Often a bug will manifest itself as a massive number of contradictions.
;; By finding all the nogooods (i.e., justifications of tms::*contra-node*) that
;; a given node participates in, one has the information that should, in theory,
;; allow the bug to be tracked.

(defun nogood-justs-including (node-or-fact)
  (let ((node (get-tms-node node-or-fact)))
    ;; NOTE: THIS WILL LOSE IF MULTIPLE CONTRA NODES
    (remove-if-not #'(lambda (just)
		       (member node (cdr just)))
		   (tms::node-justifications tms::*contra-node*))))


;;;; Checking the results of a justification
;;
;; Given the tools so far, it is easy to get access to a node and 
;; see the input environments.  Often one also needs to know WHICH
;; justification is contributing to the label of the node in question.
;; The function below provides that service.

(defun justification-result (fact n)
 (let* ((node (get-tms-node fact))
	(justs (tms::node-justifications node)))
   (cond ((< n 0) (format t "~%   Sorry, N must be positive."))
	 ((> n (length justs))
	  (format t "~%    ~A only has ~D justifications."
		  fact (length justs)))
	 (t (compute-justification-result (nth (1- n) justs))))))

(defun compute-justification-result (just)
  ;; ignore informant, just go down the nodes unioning them
  ;; together.  There may be room for another utility that weeds
  ;; out the subset of this result which is subsumed by some other
  ;; contribution.
  (do ((nodes (cddr just) (cdr nodes))
       (result (tms::label (cadr just))))
      ((null nodes) result)
    (setq result (tms::union-envss result
				  (tms::label (car nodes))))))
