;;; -*- Mode: LISP; Syntax: Common-Lisp; Package: ADB; -*-

(in-package 'adb)

;;;; Generally useful procedures

(export '(debug-flag set-debug-flag if-debug-flag defDebugFlag scratchout
	  first-order-sublis same-elements? subset-of? common-elements? generate-name
	  get-alist-value get-subindex-alist-value put-alist-value add-alist-value
	  append-alist-value append-carefully-alist-value replace-alist-value
	  splice-from-alist invert-alist-subindex invert-values-alist invert-alist
	  nested-alist-lookup nested-alist-insert ordered-cons order-list 
	  order-list-if-necessary list-in-order? ordered-merge ordered-insert
	  order-preserving-union order-preserving-intersection
	  sort-into-equivalence-classes generate-all-combinations generate-all-distinct-pairs
	  generate-power-set REP-between-files expand-between-files map-file))

(proclaim '(special *debug-flag-registry*))

;; Upward compatability crock.  Really should do something
;; more sophisticated.

(defun alphalessp (x y) (string< (format nil "~A" x) (format nil "~A" y)))

;;;; Debugging flags

(proclaim '(special *DEBUG-TABLE*))

(defun debug-flag (flag)
  (and (boundp '*DEBUG-TABLE*) (member flag *DEBUG-TABLE*)))

(defun set-debug-flag (flag val)
  (if (not (boundp '*DEBUG-TABLE*)) (setq *DEBUG-TABLE* nil))
  (if val (unless (member flag *DEBUG-TABLE*)
	    (push flag *DEBUG-TABLE*))
      (setq *DEBUG-TABLE* (delete flag *DEBUG-TABLE* ))))

(defmacro IF-DEBUG-FLAG (flag-name &rest body)
  `(and (debug-flag ',flag-name) (progn ,@ body)))

(defvar *debug-flag-registry* nil) ;;Alist of flag names and documentation strings

(defmacro defDebugFlag (name doc)
  ;; Sets up gensym for menu purposes.  Value of T means the flag is on.
  `(progn (let ((old-entry (assoc ',name *debug-flag-registry*)))
	    (when old-entry (setq *debug-flag-registry*
				  (delete old-entry *debug-flag-registry* )))
	    (push (list ',name ,doc (gensym)) *debug-flag-registry*))))

#+Symbolics
(defun menu-set-debug-flags (&aux result)
  (setq *debug-table* nil)
  (setq result (tv:choose-variable-values
		 (mapcar #'(lambda (entry)
			     (let ((var (third entry)))
			       (if (not (boundp var)) (set var nil))
			     `(,var ,(si:string (second entry))
			       :assoc (("Yes" . T)
				       ("No" . nil))))) *debug-flag-registry*)
		 :label (si:string "Debug flag settings")))
  (dolist (entry *debug-flag-registry*)
    (if (eval (third entry)) (push (first entry) *debug-table*)))
  *debug-table*)

;;;; List hacking

(defun scratchout (l1 l2) ;non-destructive and order-preserving
  (dolist (el1 l1) (setq l2 (remove el1 l2)))
  l2)

;; Substitution

(defun first-order-sublis (alist form &optional (arglist? nil) &aux car cdr)
  ;avoids substituting for the first place in an expression
    (COND ((SYMBOLP FORM)
	   (COND ((SETQ CAR (assoc FORM ALIST))
		  (CDR CAR))
		 (T FORM)))
	  ((LISTP FORM)
	   (SETQ CAR (if (and (symbolp (car form))
			      (cdr form)
			      (not arglist?)) ;assume no zero-argument functions
			 (car form)
			 (FIRST-ORDER-SUBLIS ALIST (CAR FORM) NIL))
		 CDR (FIRST-ORDER-SUBLIS ALIST (CDR FORM) T))
	   (COND ((AND (EQ (CAR FORM) CAR)
		       (EQ (CDR FORM) CDR))
		  FORM)
		 (T (CONS CAR CDR))))
	  (T FORM)))

;;;; Set operations

(defun same-elements? (l1 l2)
(do ((ll l1 (cdr ll))
     (mismatch? NIL)
     (remainder (copy-list l2)))
    ((or mismatch? (null ll)) (and (not mismatch?) (null remainder)))
  (cond ((member (car ll) l2 :test #'equal)
	 (setq remainder 
	       (delete (car ll) remainder :test #'equal)))
	(t (setq mismatch? T))))) ;assumes no dups in l1

(defun subset-of? (s1 s2)
  (do ((el1 (car s1) (car els1))
       (els1 (cdr s1) (cdr els1))
       (loser? nil))
      ((or loser? (null el1)) ;nil cannot be an element of the set
       (not loser?))
    (unless (member el1 s2 :test #'equal) (setq loser? T))))

(defun common-elements? (s1 s2)
  (dolist (el s1)
    (if (member el s2 :test #'equal) (return t))))

;;; Making named symbols

(defun generate-name (prefix &optional (package *package*) &aux index name)
  (setq index (get prefix 'generate-name-index))
  (if (null index) (setq index -1))
  (setq name (intern (format nil "~A~D" prefix (incf index)) package))
  (setf (get prefix 'generate-name-index) index)
  name)

;;;; Alist manipulating utilities
;Three kinds of alists are useful
;Normal alists, whose entries are (<key> . <value>)
;Values alists, whose entries are (<key> . <values>)
;and subindex alists, whose entries are (<key> . <alist>)

(defmacro get-alist-value (key alist)
  `(cdr (assoc ,key ,alist :test #'equal)))

(defmacro get-subindex-alist-value (key subindex-key alist)
  `(cdr (assoc ,subindex-key
	       (cdr (assoc ,key ,alist :test #'equal)) :test #'equal)))

(defmacro put-alist-value (key value alist)
  ;replaces value (standard alist)
  `(let ((loc (assoc ,key ,alist :test #'equal)))
     (cond (loc (rplacd loc ,value))
	   (t (setq ,alist (cons (cons ,key ,value) ,alist))))))

(defmacro add-alist-value (key value alist)
  ;adds value (values alist)
  `(let ((loc (assoc ,key ,alist :test #'equal)))
     (cond (loc (rplacd loc (cons ,value (delete ,value (cdr loc) :test #'equal))))
	   (t (,(if (symbolp alist) 'SETQ 'SETF)
	       ,alist (cons (list ,key ,value) ,alist))))))

(defmacro append-alist-value (key value alist)
  ;adds value (values alist)
  `(let ((loc (assoc ,key ,alist :test #'equal)))
     (cond (loc (rplacd loc (append ,value (cdr loc))))
	   (t (,(if (symbolp alist) 'SETQ 'SETF)
	       ,alist (cons (cons ,key ,value) ,alist))))))

(defmacro append-carefully-alist-value (key value alist)
  ;adds value (values alist) - very expensive, but guarentees no duplicates
  `(let ((loc (assoc ,key ,alist :test #'equal)))
     (cond (loc (rplacd loc (scratch (append ,value (cdr loc)))))
	   (t (,(if (symbolp alist) 'SETQ 'SETF)
	       ,alist (cons (cons ,key ,value) ,alist))))))

(defmacro replace-alist-value (key value alist)
  `(let ((loc (assoc ,key ,alist :test #'equal)))
     (cond (loc (rplacd loc ,value))
	   (t (setq ,alist (cons (cons ,key ,value) ,alist))))))

(defun splice-from-alist (key alist)
  ;removes an entry from any kind of alist
  (cond ((null alist) nil) ;minimize consing
	((eq (caar alist) key) (cdr alist))
	(t (splice-from-alist1 key alist)
	   alist)))

(defun splice-from-alist1 (key alist)
  (cond ((null (cdr alist)) nil)
	((eq (caadr alist) key)
	 (rplacd alist (cddr alist)))
	(t (splice-from-alist1 key (cdr alist)))))

;;;; Inverting alists
(defun invert-alist-subindex (alist)
  ;inverts index and subindex keys
  (let ((result nil))
    (dolist (element alist)
      (let ((key (car element)))
	(dolist (pair (cdr element))
	  (let ((new-pair (assoc (car pair) result :test #'equal)))
	    (cond (new-pair
		   (rplacd new-pair (cons (cons key (cdr pair))
					  (cdr new-pair))))
		  (t (push (cons (car pair) (list (cons key (cdr pair))))
			   result)))))))
    result))

(defun invert-values-alist (alist)
  (let ((result nil))
    (dolist (element alist)
      (let ((key (car element)))
	(dolist (item (cdr element))
	  (add-alist-value item key result))))
    result))

(defun invert-alist (alist)
  (let ((result nil))
    (dolist (element alist)
      (put-alist-value (cdr element) (car element) result))
    result))

;;;; Nested alists
(defun nested-alist-lookup (keys index &aux val)
  (cond ((null keys) index)
	(t (setq val (assoc (car keys) index :test #'equal))
	   (if val
	       (nested-alist-lookup (cdr keys) (cdr val))
	       nil))))

(defun nested-alist-insert (keys item alist &aux ptr)
  (setq ptr (assoc (car keys) alist :test #'equal))
  (cond ((null (cdr keys))
	 (setq ptr (assoc (car keys) alist :test #'equal))
	 (if ptr (rplacd ptr item)
	     (if alist
		 (rplacd (last alist) 
			 (cons (cons (car keys) item) nil))
		 (setq alist (cons (cons (car keys) item) nil)))))
	(t (if ptr (nested-alist-insert (cdr keys) item (cdr ptr))
	       (if alist
		   (rplacd (last alist)
			   (cons (cons (car keys)
					(nested-alist-insert (cdr keys)
							     item
							     nil)) nil))
		   (setq alist (cons (cons (car keys)
					    (nested-alist-insert (cdr keys) item nil))
				     nil))))))
  alist)

(defmacro add-to-nested-alist (keys item alist)
  `(setq ,alist (nested-alist-add ,keys ,item ,alist)))

(defun nested-alist-add (keys item alist)
  (let ((ptr (assoc (car keys) alist :test #'equal)))
    (unless ptr (setq ptr (cons (car keys) nil)
		      alist (cons ptr alist)))
    (cond ((cdr keys)
	   (setf (cdr ptr) (nested-alist-add (cdr keys)
					     item
					     (cdr ptr))))
	  (t (setf (cdr ptr) (cons item (cdr ptr))))) 
    alist))

(defmacro add-to-nested-alist-unique (keys item alist)
  `(setq ,alist (nested-alist-add-unique ,keys ,item ,alist)))

(defun nested-alist-add-unique (keys item alist)
  (let ((ptr (assoc (car keys) alist :test #'equal)))
    (unless ptr (setq ptr (cons (car keys) nil)
		      alist (cons ptr alist)))
    (cond ((cdr keys)
	   (setf (cdr ptr)
		 (nested-alist-add-unique (cdr keys)
					  item
					  (cdr ptr))))
	  ((member item (cdr ptr) :test #'equal))
	  (t (setf (cdr ptr) (cons item (cdr ptr))))) 
    alist))

;;;; Ordered lists

(defun ordered-cons (new lis &optional (test (function sxhash)))
  ;; non-destructive
  (cond ((null lis) (cons new nil))
	((equal (car lis) new) lis)
	((funcall test new (car lis))
	 (cond ((equal (cadr lis) new) lis)
	       (t (cons new lis)))) 
	(t (cons (car lis) (ordered-cons1 new (cdr lis) test)))))

(defun ordered-cons1 (new old test)
  (cond ((null old) (cons new nil))
	((equal new (car old)) old) ;don't allow duplicates
	((funcall test new (car old))
	 (cons new old))
	(t	 
	 (cons (car old) (ordered-cons1 new (cdr old) test)))))

(defun order-list (lis &optional (test (function sxhash))) (sort lis test))

(defun order-list-if-necessary (lis &optional (test (function sxhash)))
  (cond ((list-in-order? lis test) lis)
	(t (order-list lis))))

(defun list-in-order? (lis order-fun)
  (do ((ll lis (cdr ll))
       (lll (cdr lis) (cdr lll))
       (out-of-order? NIL))
      ((or (null lll) out-of-order?) (not out-of-order?))
    (and (funcall order-fun (car lll) (car ll)) (setq out-of-order? T))))

(defun ordered-merge (lis1 lis2 &optional (test (function sxhash)))
  (cond ((< (length lis1) (length lis2))
	 (ordered-merge1 (reverse lis1) lis2 test))
	(t (ordered-merge1 (reverse lis2) lis1 test))))

(defun ordered-merge1 (lis1 lis2 test)
  (cond ((null lis1) lis2)
	(t (ordered-merge1 (cdr lis1)
			   (ordered-cons (car lis1) lis2 test) test))))

;;;;Arbitrary ordering

(defun ordered-insert (el order-fun ll)
  ;; Destructive version
  (cond ((null ll) (cons el nil))
	((funcall order-fun el (car ll)) (cons el ll))
	(t (do ((current (cdr ll))
		(previous ll))
	       ((null current)
		(setf (cdr previous) (cons el nil))
		ll)
	     (cond ((funcall order-fun el (car current)) ;; Fits right here
		    (setq current (cons el current))
		    (setf (cdr previous) current)
		    (return-from ORDERED-INSERT ll))
		   (t (setq current (cdr current)
			    previous (cdr previous))))))))

(defun sort-into-equivalence-classes (elements equivalence-test)
  (do ((els (cdr elements) (cdr els))
       (el (car elements) (car els))
       (classes nil))
      ((and (null el) (null els)) classes)
    (do ((cls classes (cdr cls))
	 (found? nil))
	((or found? (null cls))
	 (cond (found? (rplacd found? (cons el (cdr found?))))
	       (t (setq classes (cons (cons el nil) classes)))))
      (cond ((funcall equivalence-test el (caar cls)) (setq found? (car cls)))))))

(defun order-preserving-union (new-stuff old-stuff &key (test #'eql))
  (dolist (new (reverse new-stuff) old-stuff)
    (pushnew new old-stuff :test test)))

(defun order-preserving-intersection (set1 set2 &key (test #'eql)
					   (ordering #'<))
  ;; Assumes set1, set2 sorted by ordering.
  ;; Is very careful to build new list structure, to keep tables
  ;; unpolluted.
  (do ((result nil))
      ((or (null set1) (null set2)) (nreverse result))
    (cond ((funcall test (car set1) (car set2))
	   (push (car set1) result) (pop set1) (pop set2))
	  ((funcall ordering (car set1) (car set2)) (pop set1))
	  (t (pop set2)))))

(defun order-preserving-intersection-nuke1 (set1 set2 &key (test #'eql)
						 (ordering #'<))
  ;; this version assumes that set1 can be spliced and diced at will.
  ;; The idea is that when one is doing successive intersections, you
  ;; are basically whittling down your set of initial prospects.
  ;; Consequently, it doesn't make sense to cons up a new result list.
  (mapl #'(lambda (set1s)
	    (do () ((null set2) (setf (car set1s) nil))
	      (cond ((funcall test (car set1s) (car set2)) ;; Got one
		     (pop set2)(return nil))
		    ((funcall ordering (car set1s) (car set2)) ;; set1 is less
		     (setf (car set1s) nil) ;; nuke it
		     (return nil))
		    (t (pop set2))))) set1)
  (delete nil set1))

;;;; Combination hacking

(defun generate-all-combinations (lists)
  ;assume no duplicate elements
  (cond ((null lists) nil)
	((null (cdr lists)) (mapcar #'(lambda (x) (cons x nil)) (car lists)))
	(t (apply 'append
		  (mapcar #'(lambda (combination)
			     (mapcar #'(lambda (el)
					(cons el combination))
				     (car lists)))
			  (generate-all-combinations (cdr lists)))))))

(defun generate-all-distinct-pairs (list)
  ;assume no duplicate elements
  (cond ((null (cdr list)) nil)
	(t (append (mapcar #'(lambda (x)
			       (list (car list) x))
			   (cdr list))
		   (generate-all-distinct-pairs (cdr list))))))

(proclaim '(special **power-sets-generated**))

(defun generate-power-set (set)
  ;be sleezy for now, won't use for n > 4 anyway...
  (setq **power-sets-generated** nil)
  (push set **power-sets-generated**)
  (generate-power-set1 set)
  **power-sets-generated**)

(defun generate-power-set1 (set)
  (cond ((null set) nil)
	((null (cdr set))
	 (unless (member set **power-sets-generated** :test #'equal)
	   (push set **power-sets-generated**)))
	(t (do ((this (cdr set) (cdr this))
		(that (car set) (car this))
		(rest nil)
		(current nil))
	       ((null that))
	     (setq current (append (reverse rest) this))
	     (unless (member current **power-sets-generated** :test #'equal)
	       (push current **power-sets-generated**)
	       (generate-power-set1 current))
	     (push that rest)))))


;;;; Examining expansions

(defun REP-between-files (from-file to-file)
  (with-Open-file (in-box from-file ':direction ':input)
    (with-open-file (out-box to-file ':direction ':output)
      (let ((mark (cons nil nil)))
	(do ((form (read in-box mark)
		   (read in-box mark))
	     (num 0 (1+ num)))
	    ((eq mark form)
	     (close in-box)
	     (close out-box) num)
	  (pprint (eval form) out-box))))))

(defun expand-between-files (from-file to-file)
  (with-open-file (in-box from-file ':direction ':input)
    (with-open-file (out-box to-file ':direction ':output)
      (let ((mark (cons nil nil)))
	(do ((form (read in-box nil mark)
		   (read in-box nil mark))
	     (num 0 (1+ num)))
	    ((eq mark form)
	     (close in-box)
	     (close out-box) num)
	  (pprint (macroexpand form) out-box))))))

(defun expand1-between-files (from-file to-file)
  (with-open-file (in-box from-file ':direction ':input)
    (with-open-file (out-box to-file ':direction ':output)
      (let ((mark (cons nil nil)))
	(do ((form (read in-box mark)
		   (read in-box mark))
	     (num 0 (1+ num)))
	    ((eq mark form)
	     (close in-box)
	     (close out-box) num)
	  (terpri out-box)
	  (pprint (macroexpand-1 form) out-box))))))

(defun map-file (fun from-file)
  (with-open-file (in-box from-file :direction :input)
    (let ((mark (cons nil nil)))
      (do ((form (read in-box mark)
		 (read in-box mark))
	   (result nil))
	  ((eq form mark)
	   (close in-box)
	   (nreverse result))
	(push (funcall fun form) result)))))
