;;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: ADB; -*-

;;;; Pattern variable manipulations

(in-package 'ADB)

(export '(variable? get-variables))

(proclaim '(special *BOUND-VARS*))

(defun variable? (x) (and (symbolp x) (eql #\?
					 (aref (symbol-name x) 0))))

(defun unvariablize (x &optional (package 'USER))
  (cond ((variable? x)
	 (intern (subseq (symbol-name x) 1) package))
	(t x)))

(defun contains-variables? (form)
(cond ((null form) nil)
      ((variable? form) T)
      ((not (consp form)) nil)
      (t (or (contains-variables? (car form))
	     (contains-variables? (cdr form))))))

(defun contains-free-variable? (form)
(cond ((null form) nil)
      ((variable? form)
       (cond ((and (boundp '*bound-vars*)
		   (member form *bound-vars* :test #'EQ))
	      NIL)
	     (T T)))
      ((not (listp form)) nil)
      (t (or (contains-free-variable? (car form))
	     (contains-free-variable? (cdr form))))))

;;;; Rule utilities

(defun bound-variable? (thing)
  (and (variable? thing)
       (boundp '*bound-vars*)
       (member thing *bound-vars* :test #'eq))) ;always symbols, so EQ is fine

(defun build-rule-substitution-form (form)
(cond ((null form) nil)
      ((variable? form) form)
      ((atom form) (list 'quote form))
      ((not (listp form)) form) ;numbers, pointers, and such 
      (t (list 'CONS 
	       (build-rule-substitution-form (car form))
	       (build-rule-substitution-form (cdr form))))))

(defmacro with-pushed-variable-bindings (new-vars . rest)
`(let ((*bound-vars* (append ,new-vars (scratchout ,new-vars *bound-vars*))))
   ,@ rest))

(defun trigger-variables (form) (pattern-free-variables form))

;;;; finding variables in expressions

(defun get-variables (form)
  (labels
    ((get-variables1 (form vars-so-far)
       (cond ((null form) vars-so-far)
	     ((variable? form)
	      (cond ((member form vars-so-far :test #'equal) vars-so-far)
		    (t (cons form vars-so-far))))
	     ((not (listp form)) vars-so-far)
	     (t (get-variables1 (cdr form)
				(get-variables1 (car form) vars-so-far))))))
    (get-variables1 form nil)))

(defun contains-occurance-of (form thing) ;thing assumed atomic
  (cond ((null form) nil)
	((atom form) (eq form thing))
	((eq form thing) T)
	(t (or (contains-occurance-of (car form) thing)
	       (contains-occurance-of (cdr form) thing)))))

(defun classify-pattern-variables (pattern)
  (labels ((classify-pattern-variables1 (pattern free bound)
	     (cond ((null pattern) (values free bound))
		   ((variable? pattern)
		    (cond ((or (member pattern free :test #'eq)
			       (member pattern bound :test #'eq))
			   (values free bound))
			  ((bound-variable? pattern)
			   (values free (cons pattern bound)))
			  (t (values (cons pattern free) bound))))
		   ((not (listp pattern)) (values free bound))
		   (t (multiple-value-bind (nfree nbound)
			  (classify-pattern-variables1 (car pattern) free bound)
			(classify-pattern-variables1
			  (cdr pattern) nfree nbound))))))
    (classify-pattern-variables1 pattern nil nil)))

(defun pattern-free-variables (pattern)
  (labels ((pattern-free-vars1 (pattern vars)
	     (cond ((null pattern) vars)
		   ((variable? pattern)
		    (if (or (member pattern vars :test #'eq)
			    (member pattern *bound-vars* :test #'eq))
			vars
			(cons pattern vars)))
		   ((not (listp pattern)) vars)
		   (t (pattern-free-vars1
			(cdr pattern) (pattern-free-vars1 (car pattern) vars))))))
    (pattern-free-vars1 pattern nil)))
